#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test of the status command."""
import os

import pytest

import orion.core.cli


def test_no_experiments(setup_pickleddb_database, monkeypatch, capsys):
    """Test status with no experiments."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    assert captured == "No experiment found\n"


def test_no_version_backward_compatible(one_experiment_no_version, capsys):
    """Test status with no experiments."""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-no-version-v1
=============================
empty


"""
    assert captured == expected


def test_python_api(with_experiment_using_python_api, capsys):
    """Test status with experiments built using python api."""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


from-python-api-v1
==================
empty


"""
    assert captured == expected


def test_missing_conf_file(with_experiment_missing_conf_file, capsys):
    """Test status can handle experiments when the user script config file is missing"""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


"""
    assert captured == expected


def test_experiment_without_trials_wout_ac(one_experiment, capsys):
    """Test status with only one experiment and no trials."""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


"""
    assert captured == expected


def test_experiment_wout_success_wout_ac(single_without_success, capsys):
    """Test status with only one experiment and no successful trial."""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
status         quantity
-----------  ----------
broken                1
interrupted           1
new                   1
reserved              1
suspended             1


"""
    assert captured == expected


def test_experiment_number_same_list_status(single_without_success, capsys):
    """Test status and list command output the consistent number of experiments"""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
status         quantity
-----------  ----------
broken                1
interrupted           1
new                   1
reserved              1
suspended             1


"""
    assert captured == expected

    orion.core.cli.main(["list"])

    captured = capsys.readouterr().out

    assert captured == " test_single_exp-v1\n"


def test_experiment_w_trials_wout_ac(single_with_trials, capsys):
    """Test status with only one experiment and all trials."""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
status         quantity    min obj
-----------  ----------  ---------
broken                1
completed             1          0
interrupted           1
new                   1
reserved              1
suspended             1


"""
    assert captured == expected


def test_two_unrelated_w_trials_wout_ac(unrelated_with_trials, capsys):
    """Test two unrelated experiments, with all types of trials."""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


test_single_exp-v1
==================
status         quantity    min obj
-----------  ----------  ---------
broken                1
completed             1          0
interrupted           1
new                   1
reserved              1
suspended             1


"""

    assert captured == expected


def test_two_related_w_trials_wout_ac(family_with_trials, capsys):
    """Test two related experiments, with all types of trials."""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


  test_double_exp_child-v1
  ========================
  status         quantity
  -----------  ----------
  broken                1
  completed             1
  interrupted           1
  new                   1
  reserved              1
  suspended             1


"""

    assert captured == expected


def test_three_unrelated_wout_ac(three_experiments_with_trials, capsys):
    """Test three unrelated experiments with all types of trials."""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


  test_double_exp_child-v1
  ========================
  status         quantity
  -----------  ----------
  broken                1
  completed             1
  interrupted           1
  new                   1
  reserved              1
  suspended             1


test_single_exp-v1
==================
status         quantity    min obj
-----------  ----------  ---------
broken                1
completed             1          0
interrupted           1
new                   1
reserved              1
suspended             1


"""

    assert captured == expected


def test_three_related_wout_ac(three_family_with_trials, capsys):
    """Test three related experiments with all types of trials."""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


  test_double_exp_child-v1
  ========================
  status         quantity
  -----------  ----------
  broken                1
  completed             1
  interrupted           1
  new                   1
  reserved              1
  suspended             1


  test_double_exp_child2-v1
  =========================
  status         quantity
  -----------  ----------
  broken                1
  completed             1
  interrupted           1
  new                   1
  reserved              1
  suspended             1


"""

    assert captured == expected


def test_three_related_branch_wout_ac(three_family_branch_with_trials, capsys):
    """Test three related experiments with all types of trials."""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


  test_double_exp_child-v1
  ========================
  status         quantity
  -----------  ----------
  broken                1
  completed             1
  interrupted           1
  new                   1
  reserved              1
  suspended             1


    test_double_exp_grand_child-v1
    ==============================
    status         quantity
    -----------  ----------
    broken                1
    completed             1
    interrupted           1
    new                   1
    reserved              1
    suspended             1


"""

    assert captured == expected


def test_one_wout_trials_w_a_wout_c(one_experiment, capsys):
    """Test experiments, without trials, with --all."""
    orion.core.cli.main(["status", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
id     status    best objective
-----  --------  ----------------
empty


"""

    assert captured == expected


def test_one_w_trials_w_a_wout_c(single_with_trials, capsys):
    """Test experiment, with all trials, with --all."""
    orion.core.cli.main(["status", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
id                                status         min obj
--------------------------------  -----------  ---------
6c10cf8f7a065c9f0bb0290800ce2926  broken
d8a550d66949a115d3db3925bd93829c  completed            0
ad2f2049de1fe1a600cde6d118588461  interrupted
9cc0d5701543d6eff7a7cd95dd0681b9  new
3de6f76692d763e3ccb1422172cccc1d  reserved
1a9fc30e1bd96b26e086196112f31a69  suspended


"""

    assert captured == expected


def test_one_wout_success_w_a_wout_c(single_without_success, capsys):
    """Test experiment, without success, with --all."""
    orion.core.cli.main(["status", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
id                                status
--------------------------------  -----------
6c10cf8f7a065c9f0bb0290800ce2926  broken
ad2f2049de1fe1a600cde6d118588461  interrupted
9cc0d5701543d6eff7a7cd95dd0681b9  new
3de6f76692d763e3ccb1422172cccc1d  reserved
1a9fc30e1bd96b26e086196112f31a69  suspended


"""

    assert captured == expected


def test_two_unrelated_w_a_wout_c(unrelated_with_trials, capsys):
    """Test two unrelated experiments with --all."""
    orion.core.cli.main(["status", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
259021a55c2a5af7074a41ed6639b2f3  broken
cbb766d729294f77f0ca86ff2bf72707  completed
ca6576848f17201852225d816fb71fcc  interrupted
28097ba31dbdffc0aa265c6bc5c98b0f  new
adbe6c400cd1e667696e28fbecd000a0  reserved
5679af6c6bb54aa8042043008ab2bc1f  suspended


test_single_exp-v1
==================
id                                status         min obj
--------------------------------  -----------  ---------
6c10cf8f7a065c9f0bb0290800ce2926  broken
d8a550d66949a115d3db3925bd93829c  completed            0
ad2f2049de1fe1a600cde6d118588461  interrupted
9cc0d5701543d6eff7a7cd95dd0681b9  new
3de6f76692d763e3ccb1422172cccc1d  reserved
1a9fc30e1bd96b26e086196112f31a69  suspended


"""

    assert captured == expected


def test_two_related_w_a_wout_c(family_with_trials, capsys):
    """Test two related experiments with --all."""
    orion.core.cli.main(["status", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
259021a55c2a5af7074a41ed6639b2f3  broken
cbb766d729294f77f0ca86ff2bf72707  completed
ca6576848f17201852225d816fb71fcc  interrupted
28097ba31dbdffc0aa265c6bc5c98b0f  new
adbe6c400cd1e667696e28fbecd000a0  reserved
5679af6c6bb54aa8042043008ab2bc1f  suspended


  test_double_exp_child-v1
  ========================
  id                                status
  --------------------------------  -----------
  890b4f07685ed020f5d9e28cac9316e1  broken
  ff81ff46da5ffe6bd623fb38a06df993  completed
  78a3e60699eee1d0b9bc51a049168fce  interrupted
  13cd454155748351790525e3079fb620  new
  d1b7ecbd3621de9195a42c76defb6603  reserved
  33d6208ef03cb236a8f3b567665c357d  suspended


"""

    assert captured == expected


def test_three_unrelated_w_a_wout_c(three_experiments_with_trials, capsys):
    """Test three unrelated experiments with --all."""
    orion.core.cli.main(["status", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
259021a55c2a5af7074a41ed6639b2f3  broken
cbb766d729294f77f0ca86ff2bf72707  completed
ca6576848f17201852225d816fb71fcc  interrupted
28097ba31dbdffc0aa265c6bc5c98b0f  new
adbe6c400cd1e667696e28fbecd000a0  reserved
5679af6c6bb54aa8042043008ab2bc1f  suspended


  test_double_exp_child-v1
  ========================
  id                                status
  --------------------------------  -----------
  890b4f07685ed020f5d9e28cac9316e1  broken
  ff81ff46da5ffe6bd623fb38a06df993  completed
  78a3e60699eee1d0b9bc51a049168fce  interrupted
  13cd454155748351790525e3079fb620  new
  d1b7ecbd3621de9195a42c76defb6603  reserved
  33d6208ef03cb236a8f3b567665c357d  suspended


test_single_exp-v1
==================
id                                status         min obj
--------------------------------  -----------  ---------
6c10cf8f7a065c9f0bb0290800ce2926  broken
d8a550d66949a115d3db3925bd93829c  completed            0
ad2f2049de1fe1a600cde6d118588461  interrupted
9cc0d5701543d6eff7a7cd95dd0681b9  new
3de6f76692d763e3ccb1422172cccc1d  reserved
1a9fc30e1bd96b26e086196112f31a69  suspended


"""

    assert captured == expected


def test_three_related_w_a_wout_c(three_family_with_trials, capsys):
    """Test three related experiments with --all."""
    orion.core.cli.main(["status", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
259021a55c2a5af7074a41ed6639b2f3  broken
cbb766d729294f77f0ca86ff2bf72707  completed
ca6576848f17201852225d816fb71fcc  interrupted
28097ba31dbdffc0aa265c6bc5c98b0f  new
adbe6c400cd1e667696e28fbecd000a0  reserved
5679af6c6bb54aa8042043008ab2bc1f  suspended


  test_double_exp_child-v1
  ========================
  id                                status
  --------------------------------  -----------
  890b4f07685ed020f5d9e28cac9316e1  broken
  ff81ff46da5ffe6bd623fb38a06df993  completed
  78a3e60699eee1d0b9bc51a049168fce  interrupted
  13cd454155748351790525e3079fb620  new
  d1b7ecbd3621de9195a42c76defb6603  reserved
  33d6208ef03cb236a8f3b567665c357d  suspended


  test_double_exp_child2-v1
  =========================
  id                                status
  --------------------------------  -----------
  1c238040d6b6d8423d99a08551fe0998  broken
  2c13424a9212ab92ea592bdaeb1c13e9  completed
  a2680fbda1faa9dfb94946cf25536f44  interrupted
  abbda454d0577ded5b8e784a9d6d5abb  new
  df58aa8fd875f129f7faa84eb15ca453  reserved
  71657e86bad0f2e8b06098a64cb883b6  suspended


"""

    assert captured == expected


def test_three_related_branch_w_a_wout_c(three_family_branch_with_trials, capsys):
    """Test three related experiments in a branch with --all."""
    orion.core.cli.main(["status", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
259021a55c2a5af7074a41ed6639b2f3  broken
cbb766d729294f77f0ca86ff2bf72707  completed
ca6576848f17201852225d816fb71fcc  interrupted
28097ba31dbdffc0aa265c6bc5c98b0f  new
adbe6c400cd1e667696e28fbecd000a0  reserved
5679af6c6bb54aa8042043008ab2bc1f  suspended


  test_double_exp_child-v1
  ========================
  id                                status
  --------------------------------  -----------
  890b4f07685ed020f5d9e28cac9316e1  broken
  ff81ff46da5ffe6bd623fb38a06df993  completed
  78a3e60699eee1d0b9bc51a049168fce  interrupted
  13cd454155748351790525e3079fb620  new
  d1b7ecbd3621de9195a42c76defb6603  reserved
  33d6208ef03cb236a8f3b567665c357d  suspended


    test_double_exp_grand_child-v1
    ==============================
    id                                status
    --------------------------------  -----------
    e374d8f802aed52c07763545f46228a7  broken
    f9ee14ff9ef0b95ed7a24860731c85a9  completed
    3f7dff101490727d5fa0efeb36ca6366  interrupted
    40838d46dbf7778a3cb51b7a09118391  new
    b7860a18b2700cce4e8009cde543975c  reserved
    cd406126bc350ad82ac77c75174cc8a2  suspended


"""

    assert captured == expected


def test_two_unrelated_w_c_wout_a(unrelated_with_trials, capsys):
    """Test two unrelated experiments with --collapse."""
    orion.core.cli.main(["status", "--collapse"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


test_single_exp-v1
==================
status         quantity    min obj
-----------  ----------  ---------
broken                1
completed             1          0
interrupted           1
new                   1
reserved              1
suspended             1


"""

    assert captured == expected


def test_two_related_w_c_wout_a(family_with_trials, capsys):
    """Test two related experiments with --collapse."""
    orion.core.cli.main(["status", "--collapse"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   2
reserved              1
suspended             1


"""

    assert captured == expected


def test_three_unrelated_w_c_wout_a(three_experiments_with_trials, capsys):
    """Test three unrelated experiments with --collapse."""
    orion.core.cli.main(["status", "--collapse"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   2
reserved              1
suspended             1


test_single_exp-v1
==================
status         quantity    min obj
-----------  ----------  ---------
broken                1
completed             1          0
interrupted           1
new                   1
reserved              1
suspended             1


"""

    assert captured == expected


def test_three_related_w_c_wout_a(three_family_with_trials, capsys):
    """Test three related experiments with --collapse."""
    orion.core.cli.main(["status", "--collapse"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   3
reserved              1
suspended             1


"""

    assert captured == expected


def test_three_related_branch_w_c_wout_a(three_family_branch_with_trials, capsys):
    """Test three related experiments with --collapse."""
    orion.core.cli.main(["status", "--collapse"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   3
reserved              1
suspended             1


"""

    assert captured == expected


def test_two_unrelated_w_ac(unrelated_with_trials, capsys):
    """Test two unrelated experiments with --collapse and --all."""
    orion.core.cli.main(["status", "--collapse", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
259021a55c2a5af7074a41ed6639b2f3  broken
cbb766d729294f77f0ca86ff2bf72707  completed
ca6576848f17201852225d816fb71fcc  interrupted
28097ba31dbdffc0aa265c6bc5c98b0f  new
adbe6c400cd1e667696e28fbecd000a0  reserved
5679af6c6bb54aa8042043008ab2bc1f  suspended


test_single_exp-v1
==================
id                                status         min obj
--------------------------------  -----------  ---------
6c10cf8f7a065c9f0bb0290800ce2926  broken
d8a550d66949a115d3db3925bd93829c  completed            0
ad2f2049de1fe1a600cde6d118588461  interrupted
9cc0d5701543d6eff7a7cd95dd0681b9  new
3de6f76692d763e3ccb1422172cccc1d  reserved
1a9fc30e1bd96b26e086196112f31a69  suspended


"""

    assert captured == expected


def test_two_related_w_ac(family_with_trials, capsys):
    """Test two related experiments with --collapse and --all."""
    orion.core.cli.main(["status", "--collapse", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
259021a55c2a5af7074a41ed6639b2f3  broken
cbb766d729294f77f0ca86ff2bf72707  completed
ca6576848f17201852225d816fb71fcc  interrupted
28097ba31dbdffc0aa265c6bc5c98b0f  new
4c409da13bdc93c54f6997797c296356  new
adbe6c400cd1e667696e28fbecd000a0  reserved
5679af6c6bb54aa8042043008ab2bc1f  suspended


"""

    assert captured == expected


def test_three_unrelated_w_ac(three_experiments_with_trials, capsys):
    """Test three unrelated experiments with --collapse and --all."""
    orion.core.cli.main(["status", "--collapse", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
259021a55c2a5af7074a41ed6639b2f3  broken
cbb766d729294f77f0ca86ff2bf72707  completed
ca6576848f17201852225d816fb71fcc  interrupted
28097ba31dbdffc0aa265c6bc5c98b0f  new
4c409da13bdc93c54f6997797c296356  new
adbe6c400cd1e667696e28fbecd000a0  reserved
5679af6c6bb54aa8042043008ab2bc1f  suspended


test_single_exp-v1
==================
id                                status         min obj
--------------------------------  -----------  ---------
6c10cf8f7a065c9f0bb0290800ce2926  broken
d8a550d66949a115d3db3925bd93829c  completed            0
ad2f2049de1fe1a600cde6d118588461  interrupted
9cc0d5701543d6eff7a7cd95dd0681b9  new
3de6f76692d763e3ccb1422172cccc1d  reserved
1a9fc30e1bd96b26e086196112f31a69  suspended


"""

    assert captured == expected


def test_three_related_w_ac(three_family_with_trials, capsys):
    """Test three related experiments with --collapse and --all."""
    orion.core.cli.main(["status", "--collapse", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
259021a55c2a5af7074a41ed6639b2f3  broken
cbb766d729294f77f0ca86ff2bf72707  completed
ca6576848f17201852225d816fb71fcc  interrupted
28097ba31dbdffc0aa265c6bc5c98b0f  new
4c409da13bdc93c54f6997797c296356  new
b97518f91e006cd4a2805657c596b11c  new
adbe6c400cd1e667696e28fbecd000a0  reserved
5679af6c6bb54aa8042043008ab2bc1f  suspended


"""

    assert captured == expected


def test_three_related_branch_w_ac(three_family_branch_with_trials, capsys):
    """Test three related experiments in a branch with --collapse and --all."""
    orion.core.cli.main(["status", "--collapse", "--all"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
259021a55c2a5af7074a41ed6639b2f3  broken
cbb766d729294f77f0ca86ff2bf72707  completed
ca6576848f17201852225d816fb71fcc  interrupted
28097ba31dbdffc0aa265c6bc5c98b0f  new
4c409da13bdc93c54f6997797c296356  new
5183ee9c28601cc78c0a148a386df9f9  new
adbe6c400cd1e667696e28fbecd000a0  reserved
5679af6c6bb54aa8042043008ab2bc1f  suspended


"""

    assert captured == expected


def test_no_experiments_w_name(setup_pickleddb_database, monkeypatch, capsys):
    """Test status when --name <exp> does not exist."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(["status", "--name", "test_ghost_exp"])

    captured = capsys.readouterr().out

    assert captured == "No experiment found\n"


def test_experiment_wout_child_w_name(unrelated_with_trials, capsys):
    """Test status with the name argument and no child."""
    orion.core.cli.main(["status", "--name", "test_single_exp"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
status         quantity    min obj
-----------  ----------  ---------
broken                1
completed             1          0
interrupted           1
new                   1
reserved              1
suspended             1


"""

    assert captured == expected


def test_experiment_w_child_w_name(three_experiments_with_trials, capsys):
    """Test status with the name argument and one child."""
    orion.core.cli.main(["status", "--name", "test_double_exp"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


  test_double_exp_child-v1
  ========================
  status         quantity
  -----------  ----------
  broken                1
  completed             1
  interrupted           1
  new                   1
  reserved              1
  suspended             1


"""

    assert captured == expected


def test_experiment_w_parent_w_name(three_experiments_with_trials, capsys):
    """Test status with the name argument and one parent."""
    orion.core.cli.main(["status", "--name", "test_double_exp_child"])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp_child-v1
========================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   2
reserved              2
suspended             1


"""

    assert captured == expected


def test_experiment_same_name_wout_exv(three_experiments_same_name, capsys):
    """Test status with three experiments having the same name but different versions."""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v3
==================
empty


"""

    assert captured == expected


def test_experiment_same_name_wout_exv_w_name(three_experiments_same_name, capsys):
    """Test status with three experiments having the same name but different versions."""
    orion.core.cli.main(["status", "--name", "test_single_exp"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v3
==================
empty


"""

    assert captured == expected


def test_experiment_same_name_wout_exv_w_child_w_name(
    three_experiments_family_same_name, capsys
):
    """Test status name with two experiments having the same name and one with a child."""
    orion.core.cli.main(["status", "--name", "test_single_exp"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


  test_single_exp-v2
  ==================
  empty


  test_single_exp_child-v1
  ========================
  empty


"""

    assert captured == expected


def test_experiment_same_name_wout_exv_w_c_w_child_w_name(
    three_experiments_family_same_name, capsys
):
    """Test status name collapsed with two experiments having the same name and one with a child."""
    orion.core.cli.main(["status", "--name", "test_single_exp", "--collapse"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v2
==================
empty


"""

    assert captured == expected


def test_experiment_same_name_wout_exv_w_child(
    three_experiments_family_same_name, capsys
):
    """Test status with two experiments having the same name and one with a child."""
    orion.core.cli.main(["status"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


  test_single_exp-v2
  ==================
  empty


  test_single_exp_child-v1
  ========================
  empty


"""

    assert captured == expected


def test_experiment_same_name_w_exv(three_experiments_same_name, capsys):
    """Test status with three experiments with the same name and `--expand-verions`."""
    orion.core.cli.main(["status", "--expand-versions"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


  test_single_exp-v2
  ==================
  empty


    test_single_exp-v3
    ==================
    empty


"""

    assert captured == expected


def test_experiment_same_name_w_exv_w_child(three_experiments_family_same_name, capsys):
    """Test status with two experiments having the same name and one with a child."""
    orion.core.cli.main(["status", "--expand-versions"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


  test_single_exp-v2
  ==================
  empty


  test_single_exp_child-v1
  ========================
  empty


"""

    assert captured == expected


def test_experiment_specific_version(three_experiments_same_name, capsys):
    """Test status using `--version`."""
    orion.core.cli.main(["status", "--version", "2"])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v2
==================
empty


"""

    assert captured == expected


def test_experiment_cant_use_version(three_experiments_same_name):
    """Test status using `--version`."""
    with pytest.raises(RuntimeError) as ex:
        orion.core.cli.main(["status", "--version", "2", "--collapse"])

    assert "collapse" in str(ex.value)

    with pytest.raises(RuntimeError) as ex:
        orion.core.cli.main(["status", "--version", "2", "--expand-versions"])

    assert "expand-versions" in str(ex.value)



# -*- coding: utf-8 -*-
"""
Plotting testing support module
===============================

Plotting testing support module providing defaults, functions and mocks.

"""
# pylint: disable=protected-access


def assert_regret_plot(plot):
    """Checks the layout of a regret plot"""
    assert plot.layout.title.text == "Regret for experiment 'experiment-name'"
    assert plot.layout.xaxis.title.text == "Trials ordered by suggested time"
    assert plot.layout.yaxis.title.text == "Objective 'loss'"

    trace1 = plot.data[0]
    assert trace1.type == "scatter"
    assert trace1.name == "trials"
    assert trace1.mode == "markers"
    assert len(trace1.y) == 1
    assert not trace1.x

    trace2 = plot.data[1]
    assert trace2.type == "scatter"
    assert trace2.name == "best-to-date"
    assert trace2.mode == "lines"
    assert len(trace2.y) == 1
    assert not trace2.x


def assert_regrets_plot(plot, names, balanced=10, with_avg=False):
    """Checks the layout of a regrets plot"""
    assert plot.layout.title.text == "Average Regret"
    assert plot.layout.xaxis.title.text == "Trials ordered by suggested time"
    assert plot.layout.yaxis.title.text == "loss"

    if with_avg:
        line_plots = plot.data[::2]
        err_plots = plot.data[1::2]
    else:
        line_plots = plot.data

    assert len(line_plots) == len(names)

    for name, trace in zip(names, line_plots):
        assert trace.type == "scatter"
        assert trace.name == name
        assert trace.mode == "lines"
        if balanced:
            assert len(trace.y) == balanced
            assert len(trace.x) == balanced

    if with_avg:
        assert len(err_plots) == len(names)
        for name, trace in zip(names, err_plots):
            assert trace.fill == "toself"
            assert trace.name == name
            assert not trace.showlegend
            if balanced:
                assert len(trace.y) == balanced * 2
                assert len(trace.x) == balanced * 2


def assert_rankings_plot(plot, names, balanced=10, with_avg=False):
    """Checks the layout of a rankings plot"""
    assert plot.layout.title.text == "Average Rankings"
    assert plot.layout.xaxis.title.text == "Trials ordered by suggested time"
    assert plot.layout.yaxis.title.text == "Ranking based on loss"

    if with_avg:
        line_plots = plot.data[::2]
        err_plots = plot.data[1::2]
    else:
        line_plots = plot.data

    assert len(line_plots) == len(names)
    for name, trace in zip(names, line_plots):
        assert trace.type == "scatter"
        assert trace.name == name
        assert trace.mode == "lines"
        if balanced:
            assert len(trace.y) == balanced
            assert len(trace.x) == balanced

    if with_avg:
        assert len(err_plots) == len(names)
        for name, trace in zip(names, err_plots):
            assert trace.fill == "toself"
            assert trace.name == name
            assert not trace.showlegend
            if balanced:
                assert len(trace.y) == balanced * 2
                assert len(trace.x) == balanced * 2


def assert_lpi_plot(plot, dims):
    """Checks the layout of a LPI plot"""
    assert plot.layout.title.text == "LPI for experiment 'experiment-name'"
    assert plot.layout.xaxis.title.text == "Hyperparameters"
    assert plot.layout.yaxis.title.text == "Local Parameter Importance (LPI)"

    trace = plot.data[0]
    assert trace["x"] == tuple(dims)
    assert trace["y"][0] > trace["y"][1]
    assert len(trace["error_y"]["array"]) == len(dims)


def assert_partial_dependencies_plot(
    plot,
    dims,
    custom_colorscale=False,
    smoothing=0.85,
    n_grid_points=5,
    n_samples=4,
    log_dims=None,
):
    """Checks the layout of a partial dependencies plot"""
    if not isinstance(n_grid_points, dict):
        n_grid_points = {dim: n_grid_points for dim in dims}
    if log_dims is None:
        log_dims = {}

    def _ax_label(axis, index):
        if index == 0:
            return f"{axis}axis"

        return f"{axis}axis{index + 1}"

    def _ax_layout(axis, index):
        return plot.layout[_ax_label(axis, index)]

    assert (
        plot.layout.title.text
        == "Partial dependencies for experiment 'experiment-name'"
    )

    assert plot.layout.coloraxis.colorbar.title.text == "Objective"
    assert plot.layout.yaxis.title.text == "Objective"

    yrange = _ax_layout("y", 0).range

    def all_indices():
        return {
            j * len(dims) + i + 1 for i in range(len(dims)) for j in range(i, len(dims))
        }

    def first_column():
        return {i * len(dims) + 1 for i in range(len(dims))}

    def last_row():
        return {len(dims) * (len(dims) - 1) + i + 1 for i in range(len(dims))}

    def diagonal():
        return {i * len(dims) + i + 1 for i in range(len(dims))}

    def assert_axis_log(axis, index, name):
        axis_type = _ax_layout(axis, index).type
        if name in log_dims:
            assert axis_type == "log"
        else:
            assert axis_type != "log"

    def assert_log_x():
        x_tested = set()
        for dim_i, dim_name in enumerate(dims):
            x_index = dim_i * len(dims) + dim_i
            for row in range(dim_i, len(dims)):
                assert_axis_log("x", x_index, dim_name)
                x_tested.add(x_index + 1)
                x_index += len(dims)

        assert x_tested == all_indices()

    assert_log_x()

    def assert_shared_y_on_diagonal():
        y_tested = set()
        for dim_i, dim_name in enumerate(dims):
            # Test shared y axis across the diagonal
            y_index = dim_i * len(dims) + dim_i
            assert _ax_layout("y", y_index).range == yrange
            y_tested.add(y_index + 1)

        assert y_tested == diagonal()

    assert_shared_y_on_diagonal()

    def assert_log_y():
        y_tested = set()
        for dim_i, dim_name in enumerate(dims):
            # Test shared y axis across the diagonal
            y_index = dim_i * len(dims) + dim_i
            # Should not be log
            assert_axis_log("y", y_index, None)
            y_tested.add(y_index + 1)

            y_index = dim_i * len(dims)
            for column in range(max(dim_i, 0)):
                assert_axis_log("y", y_index, dim_name)
                y_tested.add(y_index + 1)
                y_index += 1

        assert y_tested == all_indices()

    assert_log_y()

    def assert_x_labels():
        x_tested = set()
        for dim_i, dim_name in enumerate(dims):
            x_index = len(dims) * (len(dims) - 1) + dim_i
            assert _ax_layout("x", x_index).title.text == dim_name

        assert x_tested == last_row()

    def assert_y_labels():
        y_tested = set()
        for dim_i, dim_name in enumerate(dims):
            if dim_i > 0:
                # Test lable at left of row
                y_index = dim_i * len(dims)
                assert _ax_layout("y", y_index).title.text == dim_name
                y_tested.add(y_index + 1)
            else:
                assert _ax_layout("y", 0).title.text == "Objective"
                y_tested.add(1)

        assert y_tested == first_column()

    assert_y_labels()

    # assert x_tested == {1, 4, 5, 7, 8, 9}
    # assert y_tested == {1, 4, 5, 7, 8, 9}

    if custom_colorscale:
        assert plot.layout.coloraxis.colorscale[0][1] != "rgb(247,251,255)"
    else:
        assert plot.layout.coloraxis.colorscale[0][1] == "rgb(247,251,255)"

    data = plot.data
    data_index = 0
    for x_i, x_name in enumerate(dims):

        # Test scatter mean
        assert data[data_index].mode == "lines"
        assert data[data_index].showlegend is False
        assert len(data[data_index].x) == n_grid_points[x_name]
        assert len(data[data_index].y) == n_grid_points[x_name]
        data_index += 1
        # Test scatter var
        assert data[data_index].mode == "lines"
        assert data[data_index].fill == "toself"
        assert data[data_index].showlegend is False
        assert len(data[data_index].x) == 2 * n_grid_points[x_name]
        assert len(data[data_index].y) == 2 * n_grid_points[x_name]
        data_index += 1
        # Test scatter dots
        assert data[data_index].mode == "markers"
        assert data[data_index].showlegend is False
        assert data[data_index].customdata is not None
        assert data[data_index].hovertemplate is not None
        assert len(data[data_index].x) == n_samples
        assert len(data[data_index].y) == n_samples
        data_index += 1

        for y_i in range(x_i + 1, len(dims)):
            y_name = dims[y_i]

            # Test contour
            assert data[data_index].line.smoothing == smoothing
            # To share colorscale across subplots
            assert data[data_index].coloraxis == "coloraxis"
            assert len(data[data_index].x) == n_grid_points[x_name]
            assert len(data[data_index].y) == n_grid_points[y_name]
            assert data[data_index].z.shape == (
                n_grid_points[y_name],
                n_grid_points[x_name],
            )
            data_index += 1

            # Test scatter
            assert data[data_index].mode == "markers"
            assert data[data_index].showlegend is False
            assert len(data[data_index].x) == n_samples
            assert len(data[data_index].y) == n_samples
            data_index += 1

    # Make sure we covered all data
    assert len(data) == data_index


def assert_parallel_coordinates_plot(plot, order):
    """Checks the layout of a parallel coordinates plot"""
    assert (
        plot.layout.title.text
        == "Parallel Coordinates Plot for experiment 'experiment-name'"
    )

    trace = plot.data[0]
    for i in range(len(order)):
        assert trace.dimensions[i].label == order[i]



# -*- coding: utf-8 -*-
"""
The Evolved Transformer and large-scale evolution of image classifiers
======================================================================

Implement evolution to exploit configurations with fixed resource efficiently

"""
import copy
import importlib
import logging

import numpy as np

from orion.algo.hyperband import Hyperband, HyperbandBracket

logger = logging.getLogger(__name__)

REGISTRATION_ERROR = """
Bad fidelity level {fidelity}. Should be in {budgets}.
Params: {params}
"""

SPACE_ERROR = """
EvolutionES cannot be used if space does not contain a fidelity dimension.
"""

BUDGET_ERROR = """
Cannot build budgets below max_resources;
(max: {}) - (min: {}) > (num_rungs: {})
"""


def compute_budgets(
    min_resources, max_resources, reduction_factor, nums_population, pairs
):
    """Compute the budgets used for each execution of hyperband"""
    budgets_eves = []
    if reduction_factor == 1:
        for i in range(min_resources, max_resources + 1):
            if i == min_resources:
                budgets_eves.append([(nums_population, i)])
            else:
                budgets_eves[0].append((pairs * 2, i))
    else:
        num_brackets = int(np.log(max_resources) / np.log(reduction_factor))
        budgets = []
        budgets_tab = {}  # just for display consideration
        for bracket_id in range(0, num_brackets + 1):
            bracket_budgets = []
            num_trials = int(
                np.ceil(
                    int((num_brackets + 1) / (num_brackets - bracket_id + 1))
                    * (reduction_factor ** (num_brackets - bracket_id))
                )
            )

            min_resources = max_resources / reduction_factor ** (
                num_brackets - bracket_id
            )
            for i in range(0, num_brackets - bracket_id + 1):
                n_i = int(num_trials / reduction_factor ** i)
                min_i = int(min_resources * reduction_factor ** i)
                bracket_budgets.append((n_i, min_i))

                if budgets_tab.get(i):
                    budgets_tab[i].append((n_i, min_i))
                else:
                    budgets_tab[i] = [(n_i, min_i)]

            budgets.append(bracket_budgets)

        for i in range(len(budgets[0])):
            if i == 0:
                budgets_eves.append([(nums_population, budgets[0][i][1])])
            else:
                budgets_eves[0].append((pairs * 2, budgets[0][i][1]))

    return budgets_eves


class EvolutionES(Hyperband):
    """EvolutionES formulates hyperparameter optimization as an evolution.

    For more information on the algorithm,
    see original paper at
    https://arxiv.org/pdf/1703.01041.pdf and
    https://arxiv.org/pdf/1901.11117.pdf

    Real et al. "Large-Scale Evolution of Image Classifiers"
    So et all. "The Evolved Transformer"

    Parameters
    ----------
    space: `orion.algo.space.Space`
        Optimisation space with priors for each dimension.
    seed: None, int or sequence of int
        Seed for the random number generator used to sample new trials.
        Default: ``None``
    repetitions: int
        Number of execution of Hyperband. Default is numpy.inf which means to
        run Hyperband until no new trials can be suggested.

    """

    requires_type = None
    requires_dist = None
    requires_shape = "flattened"

    def __init__(
        self,
        space,
        seed=None,
        repetitions=np.inf,
        nums_population=20,
        mutate=None,
        max_retries=1000,
    ):
        super(EvolutionES, self).__init__(space, seed=seed, repetitions=repetitions)
        pair = nums_population // 2
        mutate_ratio = 0.3
        self.nums_population = nums_population
        self.nums_comp_pairs = pair
        self.max_retries = max_retries
        self.mutate_ratio = mutate_ratio
        self.mutate = mutate
        self.nums_mutate_gene = (
            int((len(self.space.values()) - 1) * mutate_ratio)
            if int((len(self.space.values()) - 1) * mutate_ratio) > 0
            else 1
        )

        self._param_names += ["nums_population", "mutate", "max_retries"]

        self.hurdles = []

        self.population = {}
        for key in range(len(self.space)):
            if not key == self.fidelity_index:
                self.population[key] = [-1] * nums_population

        self.performance = np.inf * np.ones(nums_population)

        self.budgets = compute_budgets(
            self.min_resources,
            self.max_resources,
            self.reduction_factor,
            nums_population,
            pair,
        )

        self.brackets = [
            BracketEVES(self, bracket_budgets, 1) for bracket_budgets in self.budgets
        ]
        self.seed_rng(seed)

    @property
    def state_dict(self):
        """Return a state dict that can be used to reset the state of the algorithm."""
        state_dict = super(EvolutionES, self).state_dict
        state_dict["population"] = copy.deepcopy(self.population)
        state_dict["performance"] = copy.deepcopy(self.performance)
        state_dict["hurdles"] = copy.deepcopy(self.hurdles)
        return state_dict

    def set_state(self, state_dict):
        """Reset the state of the algorithm based on the given state_dict"""
        super(EvolutionES, self).set_state(state_dict)
        self.population = state_dict["population"]
        self.performance = state_dict["performance"]
        self.hurdles = state_dict["hurdles"]

    def _get_bracket(self, point):
        """Get the bracket of a point during observe"""
        return self.brackets[0]


class BracketEVES(HyperbandBracket):
    """Bracket of rungs for the algorithm Hyperband.

    Parameters
    ----------
    evolutiones: `evolutiones` algorithm
        The evolutiones algorithm object which this bracket will be part of.
    budgets: list of tuple
        Each tuple gives the (n_trials, resource_budget) for the respective rung.
    repetition_id: int
        The id of hyperband execution this bracket belongs to

    """

    def __init__(self, evolution_es, budgets, repetition_id):
        super(BracketEVES, self).__init__(evolution_es, budgets, repetition_id)
        self.eves = self.hyperband
        self.search_space_remove_fidelity = []
        self._candidates = {}

        if evolution_es.mutate:
            self.mutate_attr = copy.deepcopy(evolution_es.mutate)
        else:
            self.mutate_attr = {}

        function_string = self.mutate_attr.pop(
            "function", "orion.algo.mutate_functions.default_mutate"
        )
        mod_name, func_name = function_string.rsplit(".", 1)
        mod = importlib.import_module(mod_name)
        self.mutate_func = getattr(mod, func_name)

        for i in range(len(self.space.values())):
            if not i == self.eves.fidelity_index:
                self.search_space_remove_fidelity.append(i)

    @property
    def space(self):
        return self.eves.space

    @property
    def state_dict(self):
        state_dict = super(BracketEVES, self).state_dict
        state_dict["candidates"] = copy.deepcopy(self._candidates)
        return state_dict

    def set_state(self, state_dict):
        super(BracketEVES, self).set_state(state_dict)
        self._candidates = state_dict["candidates"]

    def _get_teams(self, rung_id):
        """Get the red team and blue team"""
        if self.has_rung_filled(rung_id + 1):
            return []

        rung = self.rungs[rung_id]["results"]

        population_range = (
            self.eves.nums_population
            if len(list(rung.values())) > self.eves.nums_population
            else len(list(rung.values()))
        )

        for i in range(population_range):
            for j in self.search_space_remove_fidelity:
                self.eves.population[j][i] = list(rung.values())[i][1][j]
            self.eves.performance[i] = list(rung.values())[i][0]

        population_index = list(range(self.eves.nums_population))
        red_team = self.eves.rng.choice(
            population_index, self.eves.nums_comp_pairs, replace=False
        )
        diff_list = list(set(population_index).difference(set(red_team)))
        blue_team = self.eves.rng.choice(
            diff_list, self.eves.nums_comp_pairs, replace=False
        )

        return rung, population_range, red_team, blue_team

    def _mutate_population(self, red_team, blue_team, rung, population_range, fidelity):
        """Get the mutated population and hurdles"""
        winner_list = []
        loser_list = []

        if set(red_team) != set(blue_team):
            hurdles = 0
            for i, _ in enumerate(red_team):
                winner, loser = (
                    (red_team, blue_team)
                    if self.eves.performance[red_team[i]]
                    < self.eves.performance[blue_team[i]]
                    else (blue_team, red_team)
                )

                winner_list.append(winner[i])
                loser_list.append(loser[i])
                hurdles += self.eves.performance[winner[i]]
                self._mutate(winner[i], loser[i])

            hurdles /= len(red_team)
            self.eves.hurdles.append(hurdles)

            logger.debug("Evolution hurdles are: %s", str(self.eves.hurdles))

        points = []
        nums_all_equal = [0] * population_range
        for i in range(population_range):
            point = [0] * len(self.space)
            while True:
                point = list(point)
                point[self.eves.fidelity_index] = fidelity

                for j in self.search_space_remove_fidelity:
                    point[j] = self.eves.population[j][i]

                point = self.eves.format_point(point)

                if point in points:
                    nums_all_equal[i] += 1
                    logger.debug("find equal one, continue to mutate.")
                    self._mutate(i, i)
                elif self.eves.has_suggested(point):
                    nums_all_equal[i] += 1
                    logger.debug("find one already suggested, continue to mutate.")
                    self._mutate(i, i)
                else:
                    break
                if nums_all_equal[i] > self.eves.max_retries:
                    logger.warning(
                        "Can not Evolve any more. You can make an early stop."
                    )
                    break

            if nums_all_equal[i] < self.eves.max_retries:
                points.append(point)
            else:
                logger.debug("Dropping point %s", point)

        return points, np.array(nums_all_equal)

    def get_candidates(self, rung_id):
        """Get a candidate for promotion"""
        if rung_id not in self._candidates:
            rung, population_range, red_team, blue_team = self._get_teams(rung_id)
            fidelity = self.rungs[rung_id + 1]["resources"]
            self._candidates[rung_id] = self._mutate_population(
                red_team, blue_team, rung, population_range, fidelity
            )[0]

        candidates = []
        for candidate in self._candidates[rung_id]:
            if not self.eves.has_suggested(candidate):
                candidates.append(candidate)
        return candidates

    def _mutate(self, winner_id, loser_id):
        select_genes_key_list = self.eves.rng.choice(
            self.search_space_remove_fidelity, self.eves.nums_mutate_gene, replace=False
        )
        self.copy_winner(winner_id, loser_id)
        kwargs = copy.deepcopy(self.mutate_attr)

        for i, _ in enumerate(select_genes_key_list):
            space = self.space.values()[select_genes_key_list[i]]
            old = self.eves.population[select_genes_key_list[i]][loser_id]
            new = self.mutate_func(space, self.eves.rng, old, **kwargs)
            self.eves.population[select_genes_key_list[i]][loser_id] = new

        self.eves.performance[loser_id] = -1

    def copy_winner(self, winner_id, loser_id):
        """Copy winner to loser"""
        for key in self.search_space_remove_fidelity:
            self.eves.population[key][loser_id] = self.eves.population[key][winner_id]



# -*- coding: utf-8 -*-
"""
Utility functions for formatting data
=====================================

Conversion functions between various data types used in framework's ecosystem.

"""

from orion.core.utils.flatten import flatten
from orion.core.worker.trial import Trial


def trial_to_tuple(trial, space):
    """Extract a parameter tuple from a `orion.core.worker.trial.Trial`.

    The order within the tuple is dictated by the defined
    `orion.algo.space.Space` object.
    """
    params = flatten(trial.params)
    trial_keys = set(params.keys())
    space_keys = set(space.keys())
    if trial_keys != space_keys:
        raise ValueError(
            """"
The trial {} has wrong params:
Trial params: {}
Space dims: {}""".format(
                trial.id, sorted(trial_keys), sorted(space_keys)
            )
        )
    return tuple(params[name] for name in space.keys())


def dict_to_trial(data, space):
    """Create a `orion.core.worker.trial.Trial` object from `data`,
    filling only parameter information from `data`.

    :param data: A dict representing a sample point from `space`.
    :param space: Definition of problem's domain.
    :type space: `orion.algo.space.Space`
    """
    data = flatten(data)
    params = []
    for name, dim in space.items():
        if name not in data and dim.default_value is dim.NO_DEFAULT_VALUE:
            raise ValueError(
                "Dimension {} not specified and does not have a default value.".format(
                    name
                )
            )
        value = data.get(name, dim.default_value)

        if value not in dim:
            error_msg = "Dimension {} value {} is outside of prior {}".format(
                name, value, dim.get_prior_string()
            )
            raise ValueError(error_msg)

        params.append(dict(name=dim.name, type=dim.type, value=value))
    assert len(params) == len(space)
    return Trial(params=params)


def tuple_to_trial(data, space):
    """Create a `orion.core.worker.trial.Trial` object from `data`,
    filling only parameter information from `data`.

    :param data: A tuple representing a sample point from `space`.
    :param space: Definition of problem's domain.
    :type space: `orion.algo.space.Space`
    """
    assert len(data) == len(space)
    params = []
    for i, dim in enumerate(space.values()):
        params.append(dict(name=dim.name, type=dim.type, value=data[i]))
    return Trial(params=params)


def get_trial_results(trial):
    """Format results from a `orion.core.worker.trial.Trial` using standard structures."""
    results = dict()

    lie = trial.lie
    objective = trial.objective

    if lie:
        results["objective"] = lie.value
    elif objective:
        results["objective"] = objective.value
    else:
        results["objective"] = None

    results["constraint"] = [
        result.value for result in trial.results if result.type == "constraint"
    ]
    grad = trial.gradient
    results["gradient"] = tuple(grad.value) if grad else None

    return results


def standard_param_name(name):
    """Convert parameter name to namespace format"""
    return name.lstrip("/").lstrip("-").replace("-", "_")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module running the setup command
================================

Creates a configurarion file for the database.

"""

import logging
import os

import yaml

import orion.core
from orion.core.io.database import Database
from orion.core.utils.terminal import ask_question

log = logging.getLogger(__name__)
SHORT_DESCRIPTION = "Starts the database configuration wizard"
DESCRIPTION = """
This command starts the database configuration wizard and creates a configuration file for the
database.
"""


def add_subparser(parser):
    """Return the parser that needs to be used for this command"""
    setup_parser = parser.add_parser(
        "setup", help=SHORT_DESCRIPTION, description=DESCRIPTION
    )

    setup_parser.set_defaults(func=main)

    return setup_parser


# pylint: disable = unused-argument
def main(*args):
    """Build a configuration file."""
    default_file = orion.core.DEF_CONFIG_FILES_PATHS[-1]

    if os.path.exists(default_file):
        cancel = ""
        while cancel.strip().lower() not in ["y", "n"]:
            cancel = ask_question(
                "This will overwrite {}, do you want to proceed? (y/n) ".format(
                    default_file
                ),
                "n",
            )

        if cancel.strip().lower() == "n":
            return

    # Get database type.
    _type = ask_question(
        "Enter the database",
        choice=sorted(Database.types.keys()),
        default="mongodb",
        ignore_case=True,
    ).lower()
    # Get database arguments.
    db_class = Database.types[_type]
    db_args = db_class.get_defaults()
    arg_vals = {}
    for arg_name, default_value in sorted(db_args.items()):
        arg_vals[arg_name] = ask_question(
            "Enter the database {}: ".format(arg_name), default_value
        )

    config = {"database": {"type": _type, **arg_vals}}

    print("Default configuration file will be saved at: ")
    print(default_file)

    dirs = "/".join(default_file.split("/")[:-1])
    os.makedirs(dirs, exist_ok=True)

    with open(default_file, "w") as output:
        yaml.dump(config, output, default_flow_style=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Benchmark client
=================
"""
import datetime
import logging

from orion.benchmark import Benchmark, Study
from orion.benchmark.assessment.base import BenchmarkAssessment
from orion.benchmark.task.base import BenchmarkTask
from orion.core.io.database import DuplicateKeyError
from orion.core.utils.exceptions import NoConfigurationError
from orion.storage.base import get_storage, setup_storage

logger = logging.getLogger(__name__)


def get_or_create_benchmark(
    name, algorithms=None, targets=None, storage=None, debug=False
):
    """
    Create or get a benchmark object.

    Parameters
    ----------
    name: str
        Name of the benchmark
    algorithms: list, optional
        Algorithms used for benchmark, each algorithm can be a string or dict.
    targets: list, optional
        Targets for the benchmark, each target will be a dict with two keys.

        assess: list
            Assessment objects
        task: list
            Task objects
    storage: dict, optional
        Configuration of the storage backend.
    debug: bool, optional
        If using in debug mode, the storage config is overrided with legacy:EphemeralDB.
        Defaults to False.

    Returns
    -------
    An instance of `orion.benchmark.Benchmark`
    """
    setup_storage(storage=storage, debug=debug)

    # fetch benchmark from db
    db_config = _fetch_benchmark(name)

    benchmark_id = None
    input_configure = None

    if db_config:
        if algorithms or targets:
            input_benchmark = Benchmark(name, algorithms, targets)
            input_configure = input_benchmark.configuration
        benchmark_id, algorithms, targets = _resolve_db_config(db_config)

    if not algorithms or not targets:
        raise NoConfigurationError(
            "Benchmark {} does not exist in DB, "
            "algorithms and targets space was not defined.".format(name)
        )

    benchmark = _create_benchmark(name, algorithms, targets, storage=storage)

    if input_configure and input_benchmark.configuration != benchmark.configuration:
        logger.warn(
            "Benchmark with same name is found but has different configuration, "
            "which will be used for this creation.\n{}".format(benchmark.configuration)
        )

    if benchmark_id is None:
        logger.debug("Benchmark not found in DB. Now attempting registration in DB.")
        try:
            _register_benchmark(benchmark)
            logger.debug("Benchmark successfully registered in DB.")
        except DuplicateKeyError:
            logger.info(
                "Benchmark registration failed. This is likely due to a race condition. "
                "Now rolling back and re-attempting building it."
            )
            get_or_create_benchmark(name, algorithms, targets, storage, debug)

    return benchmark


def _get_task(name, **kwargs):
    return BenchmarkTask(of_type=name, **kwargs)


def _get_assessment(name, **kwargs):
    return BenchmarkAssessment(of_type=name, **kwargs)


def _resolve_db_config(db_config):

    benchmark_id = db_config["_id"]
    algorithms = db_config["algorithms"]

    obj_targets = []
    str_targets = db_config["targets"]
    for target in str_targets:
        obj_target = {}

        assessments = target["assess"]
        obj_assessments = []
        for name, parameters in assessments.items():
            obj_assessments.append(_get_assessment(name, **parameters))
        obj_target["assess"] = obj_assessments

        tasks = target["task"]
        obj_tasks = []
        for name, parameters in tasks.items():
            obj_tasks.append(_get_task(name, **parameters))
        obj_target["task"] = obj_tasks

    obj_targets.append(obj_target)

    targets = obj_targets

    return benchmark_id, algorithms, targets


def _create_benchmark(name, algorithms, targets, storage):

    benchmark = Benchmark(name, algorithms, targets, storage)
    benchmark.setup_studies()

    return benchmark


def _create_study(benchmark, algorithms, assess, task):
    study = Study(benchmark, algorithms, assess, task)
    study.setup_experiments()

    return study


def _fetch_benchmark(name):

    if name:
        configs = get_storage().fetch_benchmark({"name": name})
    else:
        configs = get_storage().fetch_benchmark({})

    if not configs:
        return {}

    return configs[0]


def _register_benchmark(benchmark):
    benchmark.metadata["datetime"] = datetime.datetime.utcnow()
    config = benchmark.configuration
    # This will raise DuplicateKeyError if a concurrent experiment with
    # identical (name, metadata.user) is written first in the database.
    get_storage().create_benchmark(config)



# -*- coding: utf-8 -*-
"""
Utility functions for importing modules
=======================================

Conversion functions between various data types used in framework's ecosystem.

"""

import os


def load_modules_in_path(path, filter_function=None):
    """
    Load all modules inside `path` and return a list of those
    fitting the filter function.
    """
    this_module = __import__(path, fromlist=[""])
    file_path = this_module.__path__[0]

    files = list(
        map(
            lambda f: f.split(".")[0],
            filter(lambda f2: f2.endswith("py"), os.listdir(file_path)),
        )
    )

    modules = map(lambda f: __import__(path + "." + f, fromlist=[""]), files)

    if filter_function is not None:
        modules = filter(filter_function, modules)

    return list(modules)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Task for EggHolder Function
=============================
"""
import numpy

from orion.benchmark.task.base import BaseTask


class EggHolder(BaseTask):
    """`EggHolder function <http://infinity77.net/global_optimization/test_functions_nd_E.html#go_benchmark.EggHolder>`_
    as benchmark task"""

    def __init__(self, max_trials=20, dim=2):
        self.dim = dim
        super(EggHolder, self).__init__(max_trials=max_trials, dim=dim)

    def call(self, x):
        """Evaluate a n-D eggholder function."""
        x = numpy.asarray(x)

        a = numpy.square(numpy.absolute(x[:-1] - x[1:] - 47))
        b = numpy.square(numpy.absolute(0.5 * x[:-1] + x[1:] + 47))
        summands = -x[:-1] * numpy.sin(a) - (x[1:] + 47) * numpy.sin(b)

        y = numpy.sum(summands)

        return [dict(name="eggholder", type="objective", value=y)]

    def get_search_space(self):
        """Return the search space for the task objective function"""
        rspace = {"x": "uniform(-512, 512, shape={})".format(self.dim)}

        return rspace



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for :mod:`orion.benchmark.assessment`."""

import plotly
import pytest

from orion.benchmark.assessment import AverageRank, AverageResult
from orion.testing import create_experiment, create_study_experiments
from orion.testing.plotting import assert_rankings_plot, assert_regrets_plot


class TestAverageRank:
    """Test assessment AverageRank"""

    def test_creation(self):
        """Test creation"""
        ar1 = AverageRank()
        assert ar1.task_num == 1
        assert ar1.configuration == {"AverageRank": {"task_num": 1}}

        ar2 = AverageRank(task_num=5)
        assert ar2.task_num == 5
        assert ar2.configuration == {"AverageRank": {"task_num": 5}}

    def test_analysis(self, experiment_config, trial_config):
        """Test assessment plot"""
        ar1 = AverageRank()

        with create_experiment(experiment_config, trial_config, ["completed"]) as (
            _,
            experiment,
            _,
        ):
            plot = ar1.analysis("task_name", [(0, experiment)])

        assert type(plot) is plotly.graph_objects.Figure

    @pytest.mark.usefixtures("version_XYZ")
    def test_figure_layout(self, study_experiments_config):
        """Test assessment plot format"""
        ar1 = AverageRank()

        with create_study_experiments(**study_experiments_config) as experiments:
            plot = ar1.analysis("task_name", experiments)

            assert_rankings_plot(
                plot,
                [
                    list(algorithm["algorithm"].keys())[0]
                    for algorithm in study_experiments_config["algorithms"]
                ],
                balanced=study_experiments_config["max_trial"],
                with_avg=True,
            )


class TestAverageResult:
    """Test assessment AverageResult"""

    def test_creation(self):
        """Test creation"""
        ar1 = AverageResult()
        assert ar1.task_num == 1
        assert ar1.configuration == {"AverageResult": {"task_num": 1}}

        ar2 = AverageResult(task_num=5)
        assert ar2.task_num == 5
        assert ar2.configuration == {"AverageResult": {"task_num": 5}}

    def test_analysis(self, experiment_config, trial_config):
        """Test assessment plot"""
        ar1 = AverageResult()

        with create_experiment(experiment_config, trial_config, ["completed"]) as (
            _,
            experiment,
            _,
        ):
            plot = ar1.analysis("task_name", [(0, experiment)])

        assert type(plot) is plotly.graph_objects.Figure

    @pytest.mark.usefixtures("version_XYZ")
    def test_figure_layout(self, study_experiments_config):
        """Test assessment plot format"""
        ar1 = AverageResult()

        with create_study_experiments(**study_experiments_config) as experiments:
            plot = ar1.analysis("task_name", experiments)

            assert_regrets_plot(
                plot,
                [
                    list(algorithm["algorithm"].keys())[0]
                    for algorithm in study_experiments_config["algorithms"]
                ],
                balanced=study_experiments_config["max_trial"],
                with_avg=True,
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Backward compatibility utils
============================

Helper functions to support backward compatibility.

"""
import copy
import logging
import pprint

import orion.core
from orion.core.io.orion_cmdline_parser import OrionCmdlineParser

log = logging.getLogger(__name__)


def update_user_args(metadata):
    """Make sure user script is not removed from metadata"""
    if (
        "user_script" in metadata
        and metadata["user_script"] not in metadata["user_args"]
    ):
        log.debug("Updating user_args for backward compatibility")
        metadata["user_args"] = [metadata["user_script"]] + metadata["user_args"]
        log.debug(pprint.pformat(metadata["user_args"]))


def populate_priors(metadata):
    """Compute parser state and priors based on user_args and populate metadata."""
    if "user_args" not in metadata:
        return

    update_user_args(metadata)

    parser = OrionCmdlineParser(
        orion.core.config.worker.user_script_config, allow_non_existing_files=True
    )
    if "parser" in metadata:
        # To keep configs like config user_script_config
        parser.config_prefix = metadata["parser"]["config_prefix"]
    parser.parse(metadata["user_args"])

    log.debug("Updating parser for backward compatibility")
    metadata["parser"] = parser.get_state_dict()
    log.debug(pprint.pformat(metadata["parser"]))

    log.debug("Updating priors for backward compatibility")
    metadata["priors"] = dict(parser.priors)
    log.debug(pprint.pformat(metadata["priors"]))


def update_max_broken(config):
    """Set default max_broken if None (in v <= v0.1.9)"""
    if not config.get("max_broken", None):
        config["max_broken"] = orion.core.config.experiment.max_broken
        log.debug(
            "Updating max_broken for backward compatibility: %s", config["max_broken"]
        )


def populate_space(config, force_update=True):
    """Add the space definition at the root of config."""
    if "space" in config and not force_update:
        return

    populate_priors(config["metadata"])
    # Overwrite space to make sure to include changes from user_args
    if "priors" in config["metadata"]:
        log.debug("Updating space for backward compatibility")
        log.debug(pprint.pformat(config["metadata"]["priors"]))
        config["space"] = config["metadata"]["priors"]


def db_is_outdated(database):
    """Return True if the database scheme is outdated."""
    deprecated_indices = [
        ("name", "metadata.user"),
        ("name", "metadata.user", "version"),
        "name_1_metadata.user_1",
        "name_1_metadata.user_1_version_1",
    ]

    index_information = database.index_information("experiments")
    return any(index in deprecated_indices for index in index_information.keys())


def update_db_config(config):
    """Merge DB config back into storage config"""
    config.setdefault("storage", orion.core.config.storage.to_dict())
    if "database" in config:
        log.debug("Updating db config for backward compatibility")
        config["storage"] = {"type": "legacy"}
        config["storage"]["database"] = config.pop("database")
        log.debug(pprint.pformat(config["storage"]))


def get_algo_requirements(algorithm):
    """Return a dict() of requirements of the algorithm based on interface < v0.1.10"""
    if hasattr(algorithm, "requires"):
        log.warning(
            "Algorithm.requires is deprecated and will stop being supporting in v0.3."
        )
        requirements = algorithm.requires
        requirements = (
            requirements if isinstance(requirements, list) else [requirements]
        )

        log.debug("Algorithm requirements: %s", requirements)

        requirements = copy.deepcopy(requirements)

        if "linear" in requirements:
            dist_requirement = "linear"
            del requirements[requirements.index("linear")]
        else:
            dist_requirement = None

        if "flattened" in requirements:
            shape_requirement = "flattened"
            del requirements[requirements.index("flattened")]
        else:
            shape_requirement = None

        if requirements:
            assert len(requirements) == 1
            type_requirement = requirements[0]
        else:
            type_requirement = None

        requirements = dict(
            type_requirement=type_requirement,
            shape_requirement=shape_requirement,
            dist_requirement=dist_requirement,
        )

        log.debug(
            "Algorithm requirements in new format:\n%s", pprint.pformat(requirements)
        )

        return requirements

    return dict(
        type_requirement=algorithm.requires_type,
        shape_requirement=algorithm.requires_shape,
        dist_requirement=algorithm.requires_dist,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Simple example to fill db with python api"""
from orion.client import create_experiment

create_experiment(
    "hunt-python",
    space={"x": "uniform(-50,50)"},
    algorithms={"random": {"seed": 1}},
    max_trials=10,
)

create_experiment(
    "hunt-python-branch-old",
    space={"x": "uniform(-50,50)"},
    algorithms={"random": {"seed": 1}},
    branching={"branch_from": "hunt-python"},
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform functional tests to verify backward compatibility."""
import os
import shutil
import subprocess

import pytest
from pymongo import MongoClient

import orion.core.io.experiment_builder as experiment_builder
from orion.client import create_experiment
from orion.core.io.database import Database
from orion.core.io.database.mongodb import MongoDB
from orion.core.io.database.pickleddb import PickledDB
from orion.storage.base import Storage, get_storage
from orion.storage.legacy import Legacy

DIRNAME = os.path.dirname(os.path.abspath(__file__))

PYTHON_SCRIPT_PATH = os.path.join(DIRNAME, "python_api.py")
SCRIPT_PATH = os.path.join(DIRNAME, "black_box.py")
CONFIG_FILE = os.path.join(DIRNAME, "random.yaml")

# Ignore pre-0.1.6 because was on orion.core and pypi project was deleted.
with open(os.path.join(DIRNAME, "versions.txt"), "r") as f:
    VERSIONS = [version.strip() for version in f.read().split("\n") if version.strip()]


def get_branch_argument(version):
    """Get argument to branch.

    Before v0.1.8 it was --branch. From v0.1.8 and forward it is now --branch-to.
    """
    return "--branch" if version in ["0.1.6", "0.1.7"] else "--branch-to"


def has_python_api(version):
    """Whether the python api exist in given version"""
    return not version in ["0.1.6", "0.1.7"]


def clean_mongodb():
    """Clean collections."""
    client = MongoClient(username="user", password="pass", authSource="orion_test")
    database = client.orion_test
    database.experiments.drop()
    database.lying_trials.drop()
    database.trials.drop()
    database.workers.drop()
    database.resources.drop()
    client.close()


def set_env():
    """Set the env vars for the db.

    Notes
    -----
    This function expects the env var ORION_DB_TYPE to be set.

    """
    db_type = os.environ["ORION_DB_TYPE"]
    if db_type == "pickleddb":
        file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "save.pkl")
        os.environ["ORION_DB_ADDRESS"] = file_path
        if os.path.exists(file_path):
            os.remove(file_path)
    elif db_type == "mongodb":
        os.environ["ORION_DB_ADDRESS"] = "mongodb://user:pass@localhost"
        os.environ["ORION_DB_NAME"] = "orion_test"
        clean_mongodb()


def execute(command):
    """Execute a command and return the decoded stdout."""
    out = subprocess.check_output(command.split())
    return out.decode("utf-8")


def setup_virtualenv(version):
    """Create a virtualenv and install Oríon in it for the given version"""
    virtualenv_dir = get_virtualenv_dir(version)
    if os.path.exists(virtualenv_dir):
        shutil.rmtree(virtualenv_dir)
    execute("virtualenv {}".format(virtualenv_dir))
    command = "{}/bin/pip install --ignore-installed orion=={}".format(
        virtualenv_dir, version
    )
    execute(command)


def get_version(orion_script):
    """Get Oríon version for given Oríon executer's path."""
    command = "{} --version".format(orion_script)
    stdout = subprocess.check_output(command, shell=True)

    return stdout.decode("utf-8").strip("\n")


def get_virtualenv_dir(version):
    """Get virtualenv directory for given version."""
    return "version-{}".format(version)


def fill_from_cmdline_api(orion_script, version):
    """Add experiments and trials using the commandline API"""
    # TODO(v0.3.0): Adapt when removing init_only after deprecation phase
    print(
        execute(
            " ".join(
                [
                    orion_script,
                    "-vv",
                    "init_only",
                    "--name",
                    "init-cmdline",
                    "--config",
                    CONFIG_FILE,
                    SCRIPT_PATH,
                    "-x~uniform(-50,50)",
                ]
            )
        )
    )

    # TODO(v0.3.0): Adapt when removing init_only after deprecation phase
    print(
        execute(
            " ".join(
                [
                    orion_script,
                    "-vv",
                    "init_only",
                    "--name",
                    "init-cmdline",
                    get_branch_argument(version),
                    "init-cmdline-branch-old",
                    "--config",
                    CONFIG_FILE,
                ]
            )
        )
    )

    print(
        execute(
            " ".join(
                [
                    orion_script,
                    "-vv",
                    "hunt",
                    "--name",
                    "hunt-cmdline",
                    "--config",
                    CONFIG_FILE,
                    SCRIPT_PATH,
                    "-x~uniform(-50,50)",
                ]
            )
        )
    )

    print(
        execute(
            " ".join(
                [
                    orion_script,
                    "-vv",
                    "hunt",
                    "--name",
                    "hunt-cmdline",
                    get_branch_argument(version),
                    "hunt-cmdline-branch-old",
                    "--config",
                    CONFIG_FILE,
                ]
            )
        )
    )


def fill_from_python_api(python_script, version):
    """Add experiments and trials using the python API"""
    print(execute(" ".join([python_script, PYTHON_SCRIPT_PATH, version])))


@pytest.fixture(scope="class", params=VERSIONS)
def fill_db(request):
    """Add experiments and trials in DB for given version of Oríon."""
    set_env()

    version = request.param

    setup_virtualenv(version)

    orion_script = os.path.join(get_virtualenv_dir(version), "bin", "orion")
    python_script = os.path.join(get_virtualenv_dir(version), "bin", "python")

    orion_version = get_version(orion_script)
    assert orion_version == "orion {}".format(version)

    fill_from_cmdline_api(orion_script, version)
    if has_python_api(version):
        fill_from_python_api(python_script, version)

    orion_version = get_version("orion")
    assert orion_version != "orion {}".format(version)

    print(execute("orion -vv db upgrade -f"))

    return version


def null_db_instances():
    """Nullify singleton instance so that we can assure independent instantiation tests."""
    Storage.instance = None
    Legacy.instance = None
    Database.instance = None
    MongoDB.instance = None
    PickledDB.instance = None


def build_storage():
    """Build storage from scratch"""
    null_db_instances()
    experiment_builder.setup_storage()

    return get_storage()


@pytest.mark.usefixtures("fill_db")
class TestBackwardCompatibility:
    """Tests for backward compatibility between Oríon versions."""

    def test_db_upgrade(self):
        """Verify db upgrade was successful"""
        storage = build_storage()

        index_info = storage._db.index_information("experiments")
        assert set([key for key, is_unique in index_info.items() if is_unique]) == set(
            ["_id_", "name_1_version_1"]
        )

        experiments = storage.fetch_experiments({})
        assert "version" in experiments[0]
        assert "priors" in experiments[0]["metadata"]

    def test_db_test(self):
        """Verify db test command"""
        out = execute("orion db test")
        assert "Failure" not in out

    def test_list(self, fill_db):
        """Verify list command"""
        out = execute("orion list")
        assert "init-cmdline-v1" in out
        assert "init-cmdline-branch-old-v1" in out
        assert "hunt-cmdline-v1" in out
        assert "hunt-cmdline-branch-old-v1" in out

        version = fill_db
        if has_python_api(version):
            assert "hunt-python-v1" in out

    def test_status(self, fill_db):
        """Verify status command"""
        out = execute("orion status")
        assert "init-cmdline-v1" in out
        assert "init-cmdline-branch-old-v1" in out
        assert "hunt-cmdline-v1" in out
        assert "hunt-cmdline-branch-old-v1" in out

        version = fill_db
        if has_python_api(version):
            assert "hunt-python-v1" in out

    def test_info_cmdline_api(self):
        """Verify info command from commandline api"""
        out = execute("orion info --name hunt-cmdline")
        assert "name: hunt-cmdline" in out

    def test_info_python_api(self, fill_db):
        """Verify info command from python api"""
        version = fill_db
        if not has_python_api(version):
            pytest.skip("Python API not supported by {}".format(version))

        out = execute("orion info --name hunt-python")
        assert "name: hunt-python" in out

    # TODO(v0.3.0): Adapt when removing init_only after deprecation phase
    def test_init_only(self):
        """Verify init_only command"""
        print(
            execute(
                " ".join(
                    [
                        "orion",
                        "init_only",
                        "--name",
                        "init-cmdline",
                        "--branch-to",
                        "init-cmdline-branch",
                    ]
                )
            )
        )

    def test_hunt_cmdline_api(self):
        """Verify hunt command from cmdline api parent"""
        print(
            execute(
                " ".join(
                    [
                        "orion",
                        "hunt",
                        "--name",
                        "hunt-cmdline",
                        "--branch-to",
                        "hunt-cmdline-branch",
                    ]
                )
            )
        )

    def test_hunt_python_api(self, fill_db):
        """Verify hunt command from python api parent"""
        version = fill_db
        if not has_python_api(version):
            pytest.skip("Python API not supported by {}".format(version))

        def function(x):
            """Evaluate partial information of a quadratic."""
            z = x - 34.56789
            return [
                dict(
                    name="example_objective", type="objective", value=4 * z ** 2 + 23.4
                )
            ]

        exp = create_experiment(
            "hunt-python", branching={"branch-to": "hunt-python-branch"}
        )
        exp.workon(function, max_trials=10)



"""
=============
Regret curves
=============

.. hint:: 

   Conveys how quickly the algorithm found the best hyperparameters.
    
The regret is the difference between the achieved objective and the optimal objective.
Plotted as a time series, it shows how fast an optimization algorithm approaches the best
objective. An equivalent way to visualize it is using the cumulative minimums instead of the
differences. Oríon plots the cumulative minimums so that the objective can easily be read
from the y-axis.

.. autofunction:: orion.plotting.base.regret
    :noindex:

The regret plot can be executed directly from the ``experiment`` with ``plot.regret()`` as
shown in the example below.

"""
from orion.client import get_experiment

# Specify the database where the experiments are stored. We use a local PickleDB here.
storage = dict(type="legacy", database=dict(type="pickleddb", host="../db.pkl"))

# Load the data for the specified experiment
experiment = get_experiment("random-rosenbrock", storage=storage)
fig = experiment.plot.regret()
fig

#%%
# The objective of the trials is overlayed as a scatter plot under the regret curve.
# Thanks to this we can see whether the algorithm focused its optimization close to the
# optimum (if all points are close to the regret curve near the end) or if it explored far
# from it (if many points are far from the regret curve near the end). We can see in this example
# with random search that the algorithm unsurpringly randomly explored the space.
# If we plot the results from the algorithm TPE applied on the same task, we will see a very
# different pattern
# (see how the results were generated in tutorial :ref:`sphx_glr_auto_tutorials_code_1_python_api.py`).

experiment = get_experiment("tpe-rosenbrock", storage=storage)
fig = experiment.plot.regret()
fig

#%%
# You can hover your mouse over the trials to see the configuration of the corresponding trials.
# The configuration of the trials is following this format:
#
# ::
#
#    ID: <trial id>
#    value: <objective>
#    time: <time x-axis is ordered by>
#    parameters
#      <name>: <value>

#%%
# For now, the only option to customize the regret plot is ``order_by``, the sorting order
# of the trials on the x-axis. Contributions are more than welcome to increase the customizability
# of the plot!
# By default the sorting order is ``suggested``, the order the trials were suggested by the
# optimization algorithm. Other options are ``reserved``, the time the trials started being executed
# and ``completed``, the time the trials were completed.

fig = experiment.plot.regret(order_by="completed")
fig

#%%
# In this example, the order by ``suggested`` or by ``completed`` is the same,
# but parallelized experiments can lead to different order of completed trials.

# TODO: Add documentation for `regrets()`

#%%
# Finally we save the image to serve as a thumbnail for this example. See
# the guide
# :ref:`How to save <sphx_glr_auto_examples_how-tos_code_2_how_to_save.py>`
# for more information on image saving.

fig.write_image("../../docs/src/_static/regret_thumbnail.png")

# sphinx_gallery_thumbnail_path = '_static/regret_thumbnail.png'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.io.database.pickleddb`."""
import logging
import random
import uuid
from multiprocessing import Pool

import pytest
from filelock import FileLock, SoftFileLock, Timeout
from test_database import clean_db, dump_db, get_db, orion_db

from orion.core.io.database import DatabaseTimeout, DuplicateKeyError
from orion.core.io.database.ephemeraldb import EphemeralCollection
from orion.core.io.database.pickleddb import (
    PickledDB,
    _create_lock,
    find_unpickable_doc,
    find_unpickable_field,
    local_file_systems,
)


@pytest.fixture(scope="module", autouse=True)
def db_type(pytestconfig, request):
    """Return the string identifier of a PickledDB if the --mongodb option is
    not active"""
    if pytestconfig.getoption("--mongodb"):
        pytest.skip("pickleddb tests disabled")
    yield "pickleddb"


@pytest.mark.usefixtures("clean_db")
class TestReadAndWrite(object):
    """Calls to :meth:`orion.core.io.database.AbstractDB.read_and_write`."""

    def test_logging_when_getting_file_lock(self, caplog, orion_db):
        """When logging.level is ERROR, there should be no logging."""
        logging.basicConfig(level=logging.INFO)
        caplog.clear()
        caplog.set_level(logging.ERROR)
        # any operation will trigger the lock.
        orion_db.read("experiments", {"name": "supernaedo2", "metadata.user": "dendi"})

        assert "acquired on orion_db.pkl.lock" not in caplog.text


# TESTS SET


def write(field, i):
    """Write the given value to the pickled db."""
    PickledDB.instance = None
    orion_db = PickledDB(host="orion_db.pkl")
    try:
        orion_db.write("concurrent", {field: i})
    except DuplicateKeyError:
        print("dup")
        pass

    print(field, i)


@pytest.mark.usefixtures("clean_db")
class TestConcurreny(object):
    """Test concurrent operations"""

    def test_concurrent_writes(self, orion_db):
        """Test that concurrent writes all get written properly"""
        orion_db.ensure_index("concurrent", "diff")

        assert orion_db.count("concurrent", {"diff": {"$gt": -1}}) == 0

        Pool(10).starmap(write, (("diff", i) for i in range(10)))

        assert orion_db.count("concurrent", {"diff": {"$gt": -1}}) == 10

    def test_concurrent_unique_writes(self, orion_db):
        """Test that concurrent writes cannot duplicate unique fields"""
        orion_db.ensure_index("concurrent", "unique", unique=True)

        assert orion_db.count("concurrent", {"unique": 1}) == 0

        Pool(10).starmap(write, (("unique", 1) for i in range(10)))

        assert orion_db.count("concurrent", {"unique": 1}) == 1


def test_empty_file(orion_db):
    """Check that db loading can handle empty files"""
    with open(orion_db.host, "wb") as f:
        f.write(b"")
    get_db(orion_db)


def test_unpickable_error_find_document():
    """Check error messages for pickledb"""

    class UnpickableClass:
        i_am_not_pickable = None

    unpickable_doc = {
        "_id": 2,
        "a_pickable": 1,
        "b_unpickable": UnpickableClass(),
        "c_pickable": 3,
    }

    def make_pickable(uid):
        return {"_id": uid, "a_pickable": 1, "b_pickable": 2, "c_pickable": 3}

    unpickable_dict_of_dict = [make_pickable(1), unpickable_doc, make_pickable(3)]

    pickable_dict_of_dict = [make_pickable(1), make_pickable(2), make_pickable(3)]

    unpickable_collection = EphemeralCollection()
    unpickable_collection.insert_many(unpickable_dict_of_dict)

    pickable_collection = EphemeralCollection()
    pickable_collection.insert_many(pickable_dict_of_dict)

    database = {
        "pickable_collection": pickable_collection,
        "unpickable_collection": unpickable_collection,
    }

    collection, doc = find_unpickable_doc(database)
    assert (
        collection == "unpickable_collection"
    ), "should return the unpickable document"

    key, value = find_unpickable_field(doc)
    assert key == "b_unpickable", "should return the unpickable field"
    assert isinstance(value, UnpickableClass), "should return the unpickable value"


def test_query_timeout(monkeypatch, orion_db):
    """Verify that filelock.Timeout is catched and reraised as DatabaseTimeout"""
    orion_db.timeout = 0.1

    def never_acquire(self, *arg, **kwargs):
        """Do not try to acquire, raise timeout"""
        raise Timeout(self)

    monkeypatch.setattr(FileLock, "acquire", never_acquire)

    with pytest.raises(DatabaseTimeout) as exc:
        orion_db.read("whatever", {"it should": "fail"})

    assert exc.match("Could not acquire lock for PickledDB after 0.1 seconds.")


class _MockFS:
    pass


@pytest.mark.parametrize(
    "fs_type,options,file_lock_class",
    [
        ["lustre", [], SoftFileLock],
        ["lustre", ["flock"], FileLock],
        ["lustre", ["localfilelock"], SoftFileLock],
        ["lustre", ["flock", "localflock"], SoftFileLock],
        ["beegfs", [], SoftFileLock],
        ["beegfs", ["tuneUseGlobalFileLocks"], FileLock],
        ["gpfs", [], FileLock],
        ["nfs", [], SoftFileLock],
        ["idontknow", [], SoftFileLock],
    ]
    + [[fs_type, [], FileLock] for fs_type in local_file_systems],
)
def test_file_locks(monkeypatch, fs_type, options, file_lock_class):
    """Verify that the correct file lock type is used based on FS type and configuration"""

    def _get_fs(path):
        fs = _MockFS()

        choices = [str(uuid.uuid4()) for _ in range(random.randint(3, 10))] + options
        random.shuffle(choices)
        fs.opts = ",".join(choices)
        fs.fstype = fs_type
        return fs

    monkeypatch.setattr("orion.core.io.database.pickleddb._get_fs", _get_fs)

    assert isinstance(_create_lock("/whatever/the/path/is"), file_lock_class)



# -*- coding: utf-8 -*-
"""
Base Executor
=============

Base executor class for the parallelisation of experiments.

"""

from orion.core.utils import Factory


class BaseExecutor:
    """Base executor class

    Parameters
    ----------
    n_workers: int
        The number of workers the Executor should have. Depending on the backend
        it may spawn this many worker or connect to running instance that
        dispatch work to ``n_workers`` workers.

    """

    def __init__(self, n_workers, **kwargs):
        self.n_workers = n_workers

    def __getstate__(self):
        return dict(n_workers=self.n_workers)

    def __setstate__(self, state):
        self.n_workers = state["n_workers"]

    def wait(self, futures):
        """Wait for all futures to complete execution.

        Parameters
        ----------
        futures: `concurrent.futures.Futures` or equivalent interface
            The objects returned by ``submit()`` of the executor.

        """
        pass

    def submit(self, function, *args, **kwargs):
        """Submit work to the executor for asynchronous execution

        Parameters
        ----------
        function: a callable object
            A function to be executed by the executor. The function must be serializable.
        *args, **kwargs:
            Arguments for the function. The arguments must be serializable.

        """
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


# pylint: disable=too-few-public-methods,abstract-method
class Executor(BaseExecutor, metaclass=Factory):
    """Factory class to build Executors"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for :mod:`orion.algo.hyperband`."""
import copy
import hashlib
import logging

import numpy as np
import pytest

from orion.algo.hyperband import Hyperband, HyperbandBracket, compute_budgets
from orion.algo.space import Fidelity, Integer, Real, Space
from orion.testing.algo import BaseAlgoTests, phase


@pytest.fixture
def space():
    """Create a Space with a real dimension and a fidelity value."""
    space = Space()
    space.register(Real("lr", "uniform", 0, 1))
    space.register(Fidelity("epoch", 1, 9, 3))
    return space


@pytest.fixture
def budgets():
    """Return a configuration for a bracket."""
    return [(9, 1), (3, 3), (1, 9)]


@pytest.fixture
def hyperband(space):
    """Return an instance of Hyperband."""
    return Hyperband(space, repetitions=1)


@pytest.fixture
def bracket(budgets, hyperband):
    """Return a `HyperbandBracket` instance configured with `b_config`."""
    return HyperbandBracket(hyperband, budgets, 1)


@pytest.fixture
def rung_0():
    """Create fake points and objectives for rung 0."""
    points = np.linspace(0, 8, 9)
    return dict(
        n_trials=9,
        resources=1,
        results={
            hashlib.md5(str([point]).encode("utf-8")).hexdigest(): (point, (1, point))
            for point in points
        },
    )


@pytest.fixture
def rung_1(rung_0):
    """Create fake points and objectives for rung 1."""
    values = map(
        lambda v: (v[0], (3, v[0])), list(sorted(rung_0["results"].values()))[:3]
    )
    return dict(
        n_trials=3,
        resources=3,
        results={
            hashlib.md5(str([value[0]]).encode("utf-8")).hexdigest(): value
            for value in values
        },
    )


@pytest.fixture
def rung_2(rung_1):
    """Create fake points and objectives for rung 1."""
    values = map(
        lambda v: (v[0], (9, v[0])), list(sorted(rung_1["results"].values()))[:1]
    )
    return dict(
        n_trials=1,
        resources=9,
        results={
            hashlib.md5(str([value[0]]).encode("utf-8")).hexdigest(): value
            for value in values
        },
    )


def test_compute_budgets():
    """Verify proper computation of budgets on a logarithmic scale"""
    # Check typical values
    assert compute_budgets(81, 3) == [
        [(81, 1), (27, 3), (9, 9), (3, 27), (1, 81)],
        [(27, 3), (9, 9), (3, 27), (1, 81)],
        [(9, 9), (3, 27), (1, 81)],
        [(6, 27), (2, 81)],
        [(5, 81)],
    ]
    assert compute_budgets(16, 4) == [
        [(16, 1), (4, 4), (1, 16)],
        [(4, 4), (1, 16)],
        [(3, 16)],
    ]
    assert compute_budgets(16, 5) == [[(5, 3), (1, 16)], [(2, 16)]]


def force_observe(hyperband, point, results):
    # hyperband.sampled.add(hashlib.md5(str(list(point)).encode("utf-8")).hexdigest())

    hyperband.register(point, results)

    bracket = hyperband._get_bracket(point)
    id_wo_fidelity = hyperband.get_id(point, ignore_fidelity=True)
    hyperband.trial_to_brackets[id_wo_fidelity] = bracket

    hyperband.observe([point], [results])


def mock_samples(hyperband, samples):
    for bracket in hyperband.brackets:
        bracket._samples = samples


class TestHyperbandBracket:
    """Tests for the `HyperbandBracket` class."""

    def test_rungs_creation(self, bracket):
        """Test the creation of rungs for bracket 0."""
        assert len(bracket.rungs) == 3
        assert bracket.rungs[0] == dict(n_trials=9, resources=1, results=dict())
        assert bracket.rungs[1] == dict(n_trials=3, resources=3, results=dict())
        assert bracket.rungs[2] == dict(n_trials=1, resources=9, results=dict())

    def test_register(self, hyperband, bracket):
        """Check that a point is correctly registered inside a bracket."""
        bracket.hyperband = hyperband
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode("utf-8")).hexdigest()

        bracket.register(point, 0.0)

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0]["results"]
        assert (0.0, point) == bracket.rungs[0]["results"][point_hash]

    def test_bad_register(self, hyperband, bracket):
        """Check that a non-valid point is not registered."""
        bracket.hyperband = hyperband

        with pytest.raises(IndexError) as ex:
            bracket.register((55, 0.0), 0.0)

        assert "Bad fidelity level 55" in str(ex.value)

    def test_candidate_promotion(self, hyperband, bracket, rung_0):
        """Test that correct point is promoted."""
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0

        points = bracket.get_candidates(0)

        assert points[0] == (1, 0.0)

    def test_promotion_with_rung_1_hit(self, hyperband, bracket, rung_0):
        """Test that get_candidate gives us the next best thing if point is already in rung 1."""
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode("utf-8")).hexdigest()
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0
        bracket.rungs[1]["results"][point_hash] = (0.0, point)

        points = bracket.get_candidates(0)

        assert points[0] == (1, 1)

    def test_no_promotion_when_rung_full(self, hyperband, bracket, rung_0, rung_1):
        """Test that get_candidate returns `None` if rung 1 is full."""
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0
        bracket.rungs[1] = rung_1

        points = bracket.get_candidates(0)

        assert points == []

    def test_no_promotion_if_not_completed(self, hyperband, bracket, rung_0):
        """Test the get_candidate return None if trials are not completed."""
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0
        rung = bracket.rungs[0]["results"]

        # points = bracket.get_candidates(0)

        for p_id in rung.keys():
            rung[p_id] = (None, rung[p_id][1])

        with pytest.raises(AssertionError):
            bracket.get_candidates(0)

    def test_is_done(self, bracket, rung_0):
        """Test that the `is_done` property works."""
        assert not bracket.is_done

        # Actual value of the point is not important here
        bracket.rungs[2]["results"] = {"1": (1, 0.0), "2": (1, 0.0), "3": (1, 0.0)}

        assert bracket.is_done

    def test_update_rungs_return_candidate(self, hyperband, bracket, rung_1):
        """Check if a valid modified candidate is returned by update_rungs."""
        bracket.hyperband = hyperband
        bracket.rungs[1] = rung_1
        point_hash = hashlib.md5(str([0.0]).encode("utf-8")).hexdigest()

        candidates = bracket.promote(1)

        assert point_hash in bracket.rungs[1]["results"]
        assert bracket.rungs[1]["results"][point_hash] == (0.0, (3, 0.0))
        assert candidates[0][0] == 9

    def test_update_rungs_return_no_candidate(self, hyperband, bracket, rung_1):
        """Check if no candidate is returned by update_rungs."""
        bracket.hyperband = hyperband

        candidates = bracket.promote(1)

        assert candidates == []

    def test_get_point_max_resource(self, hyperband, bracket, rung_0, rung_1, rung_2):
        """Test to get the max resource R for a particular point"""
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0

        assert bracket.get_point_max_resource(point=(1, 0.0)) == 1
        assert bracket.get_point_max_resource(point=(1, 8.0)) == 1

        bracket.rungs[1] = rung_1
        assert bracket.get_point_max_resource(point=(1, 0.0)) == 3
        assert bracket.get_point_max_resource(point=(1, 8.0)) == 1

        bracket.rungs[2] = rung_2
        assert bracket.get_point_max_resource(point=(1, 0.0)) == 9
        assert bracket.get_point_max_resource(point=(1, 8.0)) == 1

    def test_repr(self, bracket, rung_0, rung_1, rung_2):
        """Test the string representation of HyperbandBracket"""
        bracket.rungs[0] = rung_0
        bracket.rungs[1] = rung_1
        bracket.rungs[2] = rung_2

        assert str(bracket) == "HyperbandBracket(resource=[1, 3, 9], repetition id=1)"


class TestHyperband:
    """Tests for the algo Hyperband."""

    def test_register(self, hyperband, bracket, rung_0, rung_1):
        """Check that a point is registered inside the bracket."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband
        bracket.rungs = [rung_0, rung_1]
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode("utf-8")).hexdigest()

        hyperband.observe([point], [{"objective": 0.0}])

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0]["results"]
        assert (0.0, point) == bracket.rungs[0]["results"][point_hash]

    def test_register_bracket_multi_fidelity(self, space):
        """Check that a point is registered inside the same bracket for diff fidelity."""
        hyperband = Hyperband(space)

        value = 50
        fidelity = 1
        point = (fidelity, value)
        point_hash = hashlib.md5(str([value]).encode("utf-8")).hexdigest()

        force_observe(hyperband, point, {"objective": 0.0})

        bracket = hyperband.brackets[0]

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0]["results"]
        assert (0.0, point) == bracket.rungs[0]["results"][point_hash]

        fidelity = 3
        point = [fidelity, value]
        point_hash = hashlib.md5(str([value]).encode("utf-8")).hexdigest()

        force_observe(hyperband, point, {"objective": 0.0})

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[1]["results"]
        assert (0.0, point) != bracket.rungs[0]["results"][point_hash]
        assert (0.0, point) == bracket.rungs[1]["results"][point_hash]

    def test_register_next_bracket(self, space):
        """Check that a point is registered inside the good bracket when higher fidelity."""
        hyperband = Hyperband(space)

        value = 50
        fidelity = 3
        point = (fidelity, value)
        point_hash = hashlib.md5(str([value]).encode("utf-8")).hexdigest()

        force_observe(hyperband, point, {"objective": 0.0})

        assert sum(len(rung["results"]) for rung in hyperband.brackets[0].rungs) == 0
        assert sum(len(rung["results"]) for rung in hyperband.brackets[1].rungs) == 1
        assert sum(len(rung["results"]) for rung in hyperband.brackets[2].rungs) == 0
        assert point_hash in hyperband.brackets[1].rungs[0]["results"]
        assert (0.0, point) == hyperband.brackets[1].rungs[0]["results"][point_hash]

        value = 51
        fidelity = 9
        point = (fidelity, value)
        point_hash = hashlib.md5(str([value]).encode("utf-8")).hexdigest()

        force_observe(hyperband, point, {"objective": 0.0})

        assert sum(len(rung["results"]) for rung in hyperband.brackets[0].rungs) == 0
        assert sum(len(rung["results"]) for rung in hyperband.brackets[1].rungs) == 1
        assert sum(len(rung["results"]) for rung in hyperband.brackets[2].rungs) == 1
        assert point_hash in hyperband.brackets[2].rungs[0]["results"]
        assert (0.0, point) == hyperband.brackets[2].rungs[0]["results"][point_hash]

    def test_register_invalid_fidelity(self, space):
        """Check that a point cannot registered if fidelity is invalid."""
        hyperband = Hyperband(space)

        value = 50
        fidelity = 2
        point = (fidelity, value)

        with pytest.raises(ValueError) as ex:
            force_observe(hyperband, point, {"objective": 0.0})

        assert "No bracket found for point" in str(ex.value)

    def test_register_not_sampled(self, space, caplog):
        """Check that a point cannot registered if not sampled."""
        hyperband = Hyperband(space)

        value = 50
        fidelity = 2
        point = (fidelity, value)

        with caplog.at_level(logging.INFO, logger="orion.algo.hyperband"):
            hyperband.observe([point], [{"objective": 0.0}])

        assert len(caplog.records) == 1
        assert "Ignoring point" in caplog.records[0].msg

    def test_register_corrupted_db(self, caplog, space):
        """Check that a point cannot registered if passed in order diff than fidelity."""
        hyperband = Hyperband(space)

        value = 50
        fidelity = 3
        point = (fidelity, value)

        force_observe(hyperband, point, {"objective": 0.0})
        assert "Point registered to wrong bracket" not in caplog.text

        fidelity = 1
        point = [fidelity, value]

        caplog.clear()
        force_observe(hyperband, point, {"objective": 0.0})
        assert "Point registered to wrong bracket" in caplog.text

    def test_suggest_new(self, monkeypatch, hyperband, bracket, rung_0, rung_1, rung_2):
        """Test that a new point is sampled."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband

        mock_samples(hyperband, [("fidelity", i) for i in range(10)])

        points = hyperband.suggest(100)

        assert points[0] == (1.0, 0)
        assert points[1] == (1.0, 1)

    def test_suggest_duplicates_between_calls(self, monkeypatch, hyperband, bracket):
        """Test that same points are not allowed in different suggest call of
        the same hyperband execution.
        """
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband

        duplicate_point = ("fidelity", 0.0)
        new_point = ("fidelity", 0.5)

        duplicate_id = hashlib.md5(str([duplicate_point]).encode("utf-8")).hexdigest()
        bracket.rungs[0]["results"] = {duplicate_id: (0.0, duplicate_point)}

        hyperband.trial_to_brackets[
            hyperband.get_id(duplicate_point, ignore_fidelity=True)
        ] = bracket

        points = [duplicate_point, new_point]

        mock_samples(hyperband, points + [("fidelity", i) for i in range(10 - 2)])

        assert hyperband.suggest(100)[0][1] == new_point[1]

    def test_suggest_duplicates_one_call(self, monkeypatch, hyperband, bracket):
        """Test that same points are not allowed in the same suggest call ofxs
        the same hyperband execution.
        """
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband

        zhe_point = [(1, 0.0), (1, 1.0), (1, 1.0), (1, 2.0)]

        mock_samples(hyperband, zhe_point * 2)
        zhe_samples = hyperband.suggest(100)

        assert zhe_samples[0][1] == 0.0
        assert zhe_samples[1][1] == 1.0
        assert zhe_samples[2][1] == 2.0

        # zhe_point =
        mock_samples(
            hyperband,
            [
                (3, 0.0),
                (3, 1.0),
                (3, 1.0),
                (3, 2.0),
                (3, 5.0),
                (3, 4.0),
            ],
        )
        hyperband.trial_to_brackets[
            hyperband.get_id((1, 0.0), ignore_fidelity=True)
        ] = bracket
        hyperband.trial_to_brackets[
            hyperband.get_id((1, 1.0), ignore_fidelity=True)
        ] = bracket
        zhe_samples = hyperband.suggest(100)
        assert zhe_samples[0][1] == 5.0
        assert zhe_samples[1][1] == 4.0

    def test_suggest_duplicates_between_execution(
        self, monkeypatch, hyperband, budgets
    ):
        """Test that sampling collisions are handled between different hyperband execution."""
        hyperband.repetitions = 2
        bracket = HyperbandBracket(hyperband, budgets, 1)
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband

        for i in range(9):
            force_observe(hyperband, (1, i), {"objective": i})

        for i in range(3):
            force_observe(hyperband, (3, i), {"objective": i})

        force_observe(hyperband, (9, 0), {"objective": 0})

        assert not hyperband.is_done

        zhe_point = [(9, 0), (9, 1), (9, 2)]

        hyperband._refresh_brackets()
        mock_samples(hyperband, zhe_point * 2)
        zhe_samples = hyperband.suggest(100)
        assert zhe_samples == [(9, 1), (9, 2)]

    def test_suggest_inf_duplicates(
        self, monkeypatch, hyperband, bracket, rung_0, rung_1, rung_2
    ):
        """Test that sampling inf collisions will return None."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband

        zhe_point = ("fidelity", 0.0)
        hyperband.trial_to_brackets[
            hyperband.get_id(zhe_point, ignore_fidelity=True)
        ] = bracket

        mock_samples(hyperband, [zhe_point] * 2)

        assert hyperband.suggest(100) == []

    def test_suggest_in_finite_cardinality(self):
        """Test that suggest None when search space is empty"""
        space = Space()
        space.register(Integer("yolo1", "uniform", 0, 5))
        space.register(Fidelity("epoch", 1, 9, 3))

        hyperband = Hyperband(space, repetitions=1)
        for i in range(6):
            force_observe(hyperband, (1, i), {"objective": i})

        assert hyperband.suggest(100) == []

    def test_suggest_promote(self, hyperband, bracket, rung_0):
        """Test that correct point is promoted and returned."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0

        points = hyperband.suggest(100)

        assert points == [(3, i) for i in range(3)]

    def test_is_filled(self, hyperband, bracket, rung_0, rung_1, rung_2):
        """Test that Hyperband bracket detects when rung is filled."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0

        rung = bracket.rungs[0]["results"]
        trial_id = next(iter(rung.keys()))
        objective, point = rung.pop(trial_id)

        assert not bracket.is_filled
        assert not bracket.has_rung_filled(0)

        rung[trial_id] = (objective, point)

        assert bracket.is_filled
        assert bracket.has_rung_filled(0)
        assert not bracket.has_rung_filled(1)
        assert not bracket.has_rung_filled(2)

        bracket.rungs[1] = rung_1

        rung = bracket.rungs[1]["results"]
        trial_id = next(iter(rung.keys()))
        objective, point = rung.pop(trial_id)

        assert bracket.is_filled  # Should depend first rung only
        assert bracket.has_rung_filled(0)
        assert not bracket.has_rung_filled(1)

        rung[trial_id] = (objective, point)

        assert bracket.is_filled  # Should depend first rung only
        assert bracket.has_rung_filled(0)
        assert bracket.has_rung_filled(1)
        assert not bracket.has_rung_filled(2)

        bracket.rungs[2] = rung_2

        rung = bracket.rungs[2]["results"]
        trial_id = next(iter(rung.keys()))
        objective, point = rung.pop(trial_id)

        assert bracket.is_filled  # Should depend first rung only
        assert bracket.has_rung_filled(0)
        assert bracket.has_rung_filled(1)
        assert not bracket.has_rung_filled(2)

        rung[trial_id] = (objective, point)

        assert bracket.is_filled  # Should depend first rung only
        assert bracket.has_rung_filled(0)
        assert bracket.has_rung_filled(1)
        assert bracket.has_rung_filled(2)

    def test_is_ready(self, hyperband, bracket, rung_0, rung_1, rung_2):
        """Test that Hyperband bracket detects when rung is ready."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband
        bracket.rungs[0] = rung_0

        rung = bracket.rungs[0]["results"]
        trial_id = next(iter(rung.keys()))
        objective, point = rung[trial_id]
        rung[trial_id] = (None, point)

        assert not bracket.is_ready()
        assert not bracket.is_ready(0)

        rung[trial_id] = (objective, point)

        assert bracket.is_ready()
        assert bracket.is_ready(0)
        assert not bracket.is_ready(1)
        assert not bracket.is_ready(2)

        bracket.rungs[1] = rung_1

        rung = bracket.rungs[1]["results"]
        trial_id = next(iter(rung.keys()))
        objective, point = rung[trial_id]
        rung[trial_id] = (None, point)

        assert not bracket.is_ready()  # Should depend on last rung that contains trials
        assert bracket.is_ready(0)
        assert not bracket.is_ready(1)
        assert not bracket.is_ready(2)

        rung[trial_id] = (objective, point)

        assert bracket.is_ready()  # Should depend on last rung that contains trials
        assert bracket.is_ready(0)
        assert bracket.is_ready(1)
        assert not bracket.is_ready(2)

        bracket.rungs[2] = rung_2

        rung = bracket.rungs[2]["results"]
        trial_id = next(iter(rung.keys()))
        objective, point = rung[trial_id]
        rung[trial_id] = (None, point)

        assert not bracket.is_ready()  # Should depend on last rung that contains trials
        assert bracket.is_ready(0)
        assert bracket.is_ready(1)
        assert not bracket.is_ready(2)

        rung[trial_id] = (objective, point)

        assert bracket.is_ready()  # Should depend on last rung that contains trials
        assert bracket.is_ready(0)
        assert bracket.is_ready(1)
        assert bracket.is_ready(2)

    def test_suggest_opt_out(self, hyperband, bracket, rung_0, rung_1, rung_2):
        """Test that Hyperband opts out when rungs are not ready."""
        hyperband.brackets = [bracket]
        bracket.hyperband = hyperband

        bracket.rungs[0] = rung_0

        trial_id = next(iter(rung_0["results"].keys()))
        objective, point = rung_0["results"][trial_id]
        rung_0["results"][trial_id] = (None, point)

        assert hyperband.suggest(100) == []

    def test_full_process(self, monkeypatch, hyperband):
        """Test Hyperband full process."""
        sample_points = [("fidelity", i) for i in range(100)]

        hyperband._refresh_brackets()
        mock_samples(hyperband, copy.deepcopy(sample_points))

        # Fill all brackets' first rung

        points = hyperband.suggest(100)
        from orion.algo.hyperband import tabulate_status

        print(tabulate_status(hyperband.brackets))
        assert points[:3] == [(9, i) for i in range(3)]
        assert points[3:6] == [(3, i) for i in range(3, 6)]
        assert points[6:] == [(1, i) for i in range(6, 15)]

        assert hyperband.brackets[0].has_rung_filled(0)
        assert not hyperband.brackets[0].is_ready()
        assert hyperband.suggest(100) == []
        assert hyperband.suggest(100) == []

        # Observe first bracket first rung

        for i in range(9):
            hyperband.observe([(1, i + 3 + 3)], [{"objective": 16 - i}])

        assert hyperband.brackets[0].is_ready()
        assert not hyperband.brackets[1].is_ready()
        assert not hyperband.brackets[2].is_ready()

        # Promote first bracket first rung
        points = hyperband.suggest(100)
        assert points == [(3, 3 + 3 + 9 - 1 - i) for i in range(3)]

        assert hyperband.brackets[0].has_rung_filled(1)
        assert not hyperband.brackets[0].is_ready()
        assert not hyperband.brackets[1].is_ready()
        assert not hyperband.brackets[2].is_ready()

        # Observe first bracket second rung
        for i in range(3):
            hyperband.observe([(3, 3 + 3 + 9 - 1 - i)], [{"objective": 8 - i}])

        assert hyperband.brackets[0].is_ready()
        assert not hyperband.brackets[1].is_ready()
        assert not hyperband.brackets[2].is_ready()

        # Promote first bracket second rung
        points = hyperband.suggest(100)
        assert points == [(9, 12)]

        assert hyperband.brackets[0].has_rung_filled(2)
        assert not hyperband.brackets[0].is_ready()
        assert not hyperband.brackets[1].is_ready()
        assert not hyperband.brackets[2].is_ready()

        # Observe second bracket first rung
        for i in range(3):
            hyperband.observe([(3, i + 3)], [{"objective": 8 - i}])

        assert not hyperband.brackets[0].is_ready()
        assert hyperband.brackets[1].is_ready()
        assert not hyperband.brackets[2].is_ready()

        # Promote second bracket first rung
        points = hyperband.suggest(100)
        assert points == [(9, 5)]

        assert not hyperband.brackets[0].is_ready()
        assert hyperband.brackets[1].has_rung_filled(1)
        assert not hyperband.brackets[1].is_ready()
        assert not hyperband.brackets[2].is_ready()

        # Observe third bracket first rung
        for i in range(3):
            hyperband.observe([(9, i)], [{"objective": 3 - i}])

        assert not hyperband.brackets[0].is_ready(2)
        assert not hyperband.brackets[1].is_ready(1)
        assert hyperband.brackets[2].is_ready(0)
        assert hyperband.brackets[2].is_done

        # Observe second bracket second rung
        for i in range(1):
            hyperband.observe([(9, 3 + 3 - 1 - i)], [{"objective": 5 - i}])

        assert not hyperband.brackets[0].is_ready(2)
        assert hyperband.brackets[1].is_ready(1)
        assert hyperband.brackets[1].is_done

        # Observe first bracket third rung
        hyperband.observe(points, [{"objective": 3 - i}])

        assert hyperband.is_done
        assert hyperband.brackets[0].is_done
        assert hyperband.suggest(100) == []

        # Refresh repeat and execution times property
        monkeypatch.setattr(hyperband, "repetitions", 2)
        # monkeypatch.setattr(hyperband.brackets[0], "repetition_id", 0)
        # hyperband.observe([(9, 12)], [{"objective": 3 - i}])
        hyperband._refresh_brackets()
        mock_samples(hyperband, copy.deepcopy(sample_points))
        points = hyperband.suggest(100)
        assert not hyperband.is_done
        assert not hyperband.brackets[0].is_ready(2)
        assert not hyperband.brackets[0].is_done
        assert points[:3] == [(9, 3), (9, 4), (9, 6)]
        assert points[3:6] == [(3, 7), (3, 8), (3, 9)]
        assert points[6:] == [(1, i) for i in range(15, 24)]


class TestGenericHyperband(BaseAlgoTests):
    algo_name = "hyperband"
    config = {
        "seed": 123456,
        "repetitions": 3,
    }
    space = {"x": "uniform(0, 1)", "f": "fidelity(1, 10, base=2)"}

    @phase
    def test_suggest_lots(self, mocker, num, attr):
        """Test that hyperband returns whole rungs when requesting large `num`"""
        algo = self.create_algo()
        spy = self.spy_phase(mocker, num, algo, attr)
        points = algo.suggest(10000)
        repetition_id, rung_id = self.infer_repetition_and_rung(num)
        assert len(points) == BUDGETS[rung_id + 1 if rung_id < 3 else 0]

    def test_suggest_n(self, mocker, num, attr):
        algo = self.create_algo()
        spy = self.spy_phase(mocker, num, algo, attr)
        points = algo.suggest(5)
        repetition_id, rung_id = self.infer_repetition_and_rung(num)
        assert len(points) == min(BUDGETS[rung_id + 1 if rung_id < 3 else 0], 5)

    @pytest.mark.skip(reason="See https://github.com/Epistimio/orion/issues/598")
    def test_is_done_cardinality(self):
        space = self.update_space(
            {
                # Increase fidelity to increase number of trials in first rungs
                "f": "fidelity(1, 100, base=2)",
                "x": "uniform(0, 4, discrete=True)",
                "y": "choices(['a', 'b', 'c'])",
                "z": "loguniform(1, 6, discrete=True)",
            }
        )
        space = self.create_space(space)
        assert space.cardinality == 5 * 3 * 6

        algo = self.create_algo(space=space)

        for rung in range(len(algo.algorithm.brackets[0].rungs)):
            assert not algo.is_done

            points = []
            assert not algo.is_done
            n_sampled = algo.n_suggested
            n_trials = len(algo.algorithm.trial_to_brackets)
            new_points = algo.suggest(100)
            if not new_points:
                break
            points += new_points
            assert algo.n_suggested == n_sampled + len(new_points)
            assert len(algo.algorithm.trial_to_brackets) == space.cardinality

            # We reached max number of trials we can suggest before observing any.
            assert algo.suggest(100) == []

            assert not algo.is_done

            for i, point in enumerate(points):
                algo.observe([point], [dict(objective=i)])

        assert algo.is_done

    @pytest.mark.parametrize("num", [100000, 1])
    def test_is_done_max_trials(self, num):
        space = self.create_space()

        MAX_TRIALS = 10
        algo = self.create_algo(space=space)
        algo.algorithm.max_trials = MAX_TRIALS

        objective = 0
        while not algo.is_done:
            points = algo.suggest(num)
            assert points is not None
            if points:
                self.observe_points(points, algo, objective)
                objective += len(points)

        # Hyperband should ignore max trials.
        assert algo.n_observed > MAX_TRIALS
        assert algo.is_done

    @pytest.mark.skip(reason="See https://github.com/Epistimio/orion/issues/599")
    def test_optimize_branin(self):
        pass

    def infer_repetition_and_rung(self, num):
        budgets = list(np.cumsum(BUDGETS))
        if num >= budgets[-1] * 2:
            return 3, -1
        elif num >= budgets[-1]:
            return 2, -1

        if num <= 1:
            return 1, -1

        return 1, budgets.index(num)

    def assert_callbacks(self, spy, num, algo):
        if num == 0:
            return

        repetition_id, rung_id = self.infer_repetition_and_rung(num)

        brackets = algo.algorithm.brackets

        assert len(brackets) == repetition_id * len(brackets[0].rungs)

        for j in range(0, rung_id + 1):
            for bracket in brackets:
                if len(bracket.rungs) > j:
                    assert len(bracket.rungs[j]["results"]) > 0, (bracket, j)


BUDGETS = [20, 8, 3, 1]

TestGenericHyperband.set_phases(
    [("random", 0, "space.sample")]
    + [
        (f"rung{i}", budget, "suggest")
        for i, budget in enumerate(np.cumsum(BUDGETS[:-1]))
    ]
    + [("rep1-rung1", sum(BUDGETS), "suggest")]
    + [("rep2-rung1", sum(BUDGETS) * 2, "suggest")]
)



# -*- coding: utf-8 -*-
"""
Legacy storage
==============

Old Storage implementation.

"""
import datetime
import json
import logging

import orion.core
import orion.core.utils.backward as backward
from orion.core.io.database import Database, OutdatedDatabaseError
from orion.core.utils.exceptions import MissingResultFile
from orion.core.worker.trial import Trial, validate_status
from orion.storage.base import (
    BaseStorageProtocol,
    FailedUpdate,
    MissingArguments,
    get_uid,
)

log = logging.getLogger(__name__)


def get_database():
    """Return current database

    This is a wrapper around the Database Singleton object to provide
    better error message when it is used without being initialized.

    Raises
    ------
    RuntimeError
        If the underlying database was not initialized prior to calling this function

    Notes
    -----
    To initialize the underlying database you must first call `Database(...)`
    with the appropriate arguments for the chosen backend

    """
    return Database()


def setup_database(config=None):
    """Create the Database instance from a configuration.

    Parameters
    ----------
    config: dict
        Configuration for the database backend. If not defined, global configuration
        is used.

    """
    if config is None:
        # TODO: How could we support orion.core.config.storage.database as well?
        config = orion.core.config.database.to_dict()

    db_opts = config
    dbtype = db_opts.pop("type")

    log.debug("Creating %s database client with args: %s", dbtype, db_opts)
    try:
        Database(of_type=dbtype, **db_opts)
    except ValueError:
        if Database().__class__.__name__.lower() != dbtype.lower():
            raise


class Legacy(BaseStorageProtocol):
    """Legacy protocol, store all experiments and trials inside the Database()

    Parameters
    ----------
    config: Dict
        configuration definition passed from experiment_builder
        to storage factory to legacy constructor.
        See `~orion.io.database.Database` for more details
    setup: bool
        Setup the database (create indexes)

    """

    def __init__(self, database=None, setup=True):
        if database is not None:
            setup_database(database)

        self._db = Database()

        if setup:
            self._setup_db()

    def _setup_db(self):
        """Database index setup"""
        if backward.db_is_outdated(self._db):
            raise OutdatedDatabaseError(
                "The database is outdated. You can upgrade it with the "
                "command `orion db upgrade`."
            )

        self._db.index_information("experiment")
        self._db.ensure_index(
            "experiments",
            [("name", Database.ASCENDING), ("version", Database.ASCENDING)],
            unique=True,
        )

        self._db.ensure_index("experiments", "metadata.datetime")

        self._db.ensure_index("benchmarks", "name", unique=True)

        self._db.ensure_index("trials", "experiment")
        self._db.ensure_index("trials", "status")
        self._db.ensure_index("trials", "results")
        self._db.ensure_index("trials", "start_time")
        self._db.ensure_index("trials", [("end_time", Database.DESCENDING)])

    def create_benchmark(self, config):
        """Insert a new benchmark inside the database"""
        return self._db.write("benchmarks", data=config, query=None)

    def fetch_benchmark(self, query, selection=None):
        """Fetch all benchmarks that match the query"""
        return self._db.read("benchmarks", query, selection)

    def create_experiment(self, config):
        """See :func:`orion.storage.base.BaseStorageProtocol.create_experiment`"""
        return self._db.write("experiments", data=config, query=None)

    def delete_experiment(self, experiment=None, uid=None):
        """See :func:`orion.storage.base.BaseStorageProtocol.delete_experiment`"""
        uid = get_uid(experiment, uid)
        return self._db.remove("experiments", query={"_id": uid})

    def update_experiment(self, experiment=None, uid=None, where=None, **kwargs):
        """See :func:`orion.storage.base.BaseStorageProtocol.update_experiment`"""
        uid = get_uid(experiment, uid)

        if where is None:
            where = dict()

        if uid is not None:
            where["_id"] = uid
        return self._db.write("experiments", data=kwargs, query=where)

    def fetch_experiments(self, query, selection=None):
        """See :func:`orion.storage.base.BaseStorageProtocol.fetch_experiments`"""
        return self._db.read("experiments", query, selection)

    def fetch_trials(self, experiment=None, uid=None, where=None):
        """See :func:`orion.storage.base.BaseStorageProtocol.fetch_trials`"""
        uid = get_uid(experiment, uid)

        if where is None:
            where = dict()

        where["experiment"] = uid

        return self._fetch_trials(where)

    def _fetch_trials(self, query, selection=None):
        """See :func:`orion.storage.base.BaseStorageProtocol.fetch_trials`"""

        def sort_key(item):
            submit_time = item.submit_time
            if submit_time is None:
                return 0
            return submit_time

        trials = Trial.build(self._db.read("trials", query=query, selection=selection))
        trials.sort(key=sort_key)

        return trials

    def register_trial(self, trial):
        """See :func:`orion.storage.base.BaseStorageProtocol.register_trial`"""
        self._db.write("trials", trial.to_dict())
        return trial

    def delete_trials(self, experiment=None, uid=None, where=None):
        """See :func:`orion.storage.base.BaseStorageProtocol.delete_trials`"""
        uid = get_uid(experiment, uid)

        if where is None:
            where = dict()

        if uid is not None:
            where["experiment"] = uid
        return self._db.remove("trials", query=where)

    def register_lie(self, trial):
        """See :func:`orion.storage.base.BaseStorageProtocol.register_lie`"""
        return self._db.write("lying_trials", trial.to_dict())

    def retrieve_result(self, trial, **kwargs):
        """Do nothing for the legacy backend.

        Trial object should already have its results at this point.

        Parameters
        ----------
        trial: Trial
            The trial object to be updated

        Returns
        -------
        returns the trial object

        Notes
        -----
        This does not update the database!

        """
        return trial

    def get_trial(self, trial=None, uid=None):
        """See :func:`orion.storage.base.BaseStorageProtocol.get_trial`"""
        if trial is not None and uid is not None:
            assert trial._id == uid

        if uid is None:
            if trial is None:
                raise MissingArguments("Either `trial` or `uid` should be set")

            uid = trial.id

        result = self._db.read("trials", {"_id": uid})
        if not result:
            return None

        return Trial(**result[0])

    def update_trials(self, experiment=None, uid=None, where=None, **kwargs):
        """See :func:`orion.storage.base.BaseStorageProtocol.update_trials`"""
        uid = get_uid(experiment, uid)
        if where is None:
            where = dict()

        where["experiment"] = uid
        return self._db.write("trials", data=kwargs, query=where)

    def update_trial(self, trial=None, uid=None, where=None, **kwargs):
        """See :func:`orion.storage.base.BaseStorageProtocol.update_trial`"""
        uid = get_uid(trial, uid)

        if where is None:
            where = dict()

        where["_id"] = uid
        return self._db.write("trials", data=kwargs, query=where)

    def fetch_lost_trials(self, experiment):
        """See :func:`orion.storage.base.BaseStorageProtocol.fetch_lost_trials`"""
        heartbeat = orion.core.config.worker.heartbeat
        threshold = datetime.datetime.utcnow() - datetime.timedelta(
            seconds=heartbeat * 5
        )
        lte_comparison = {"$lte": threshold}
        query = {
            "experiment": experiment._id,
            "status": "reserved",
            "heartbeat": lte_comparison,
        }

        return self._fetch_trials(query)

    def push_trial_results(self, trial):
        """See :func:`orion.storage.base.BaseStorageProtocol.push_trial_results`"""
        rc = self.update_trial(
            trial, **trial.to_dict(), where={"_id": trial.id, "status": "reserved"}
        )
        if not rc:
            raise FailedUpdate()

        return rc

    def set_trial_status(self, trial, status, heartbeat=None, was=None):
        """See :func:`orion.storage.base.BaseStorageProtocol.set_trial_status`"""
        heartbeat = heartbeat or datetime.datetime.utcnow()
        was = was or trial.status

        validate_status(status)
        validate_status(was)

        update = dict(status=status, heartbeat=heartbeat, experiment=trial.experiment)

        rc = self.update_trial(trial, **update, where={"status": was, "_id": trial.id})

        if not rc:
            raise FailedUpdate()

        trial.status = status

    def fetch_pending_trials(self, experiment):
        """See :func:`orion.storage.base.BaseStorageProtocol.fetch_pending_trials`"""
        query = dict(
            experiment=experiment._id,
            status={"$in": ["new", "suspended", "interrupted"]},
        )
        return self._fetch_trials(query)

    def reserve_trial(self, experiment):
        """See :func:`orion.storage.base.BaseStorageProtocol.reserve_trial`"""
        query = dict(
            experiment=experiment._id,
            status={"$in": ["interrupted", "new", "suspended"]},
        )
        # read and write works on a single document
        now = datetime.datetime.utcnow()
        trial = self._db.read_and_write(
            "trials",
            query=query,
            data=dict(status="reserved", start_time=now, heartbeat=now),
        )

        if trial is None:
            return None

        return Trial(**trial)

    def fetch_noncompleted_trials(self, experiment):
        """See :func:`orion.storage.base.BaseStorageProtocol.fetch_noncompleted_trials`"""
        query = dict(experiment=experiment._id, status={"$ne": "completed"})
        return self._fetch_trials(query)

    def count_completed_trials(self, experiment):
        """See :func:`orion.storage.base.BaseStorageProtocol.count_completed_trials`"""
        query = dict(experiment=experiment._id, status="completed")
        return self._db.count("trials", query)

    def count_broken_trials(self, experiment):
        """See :func:`orion.storage.base.BaseStorageProtocol.count_broken_trials`"""
        query = dict(experiment=experiment._id, status="broken")
        return self._db.count("trials", query)

    def update_heartbeat(self, trial):
        """Update trial's heartbeat"""
        return self.update_trial(
            trial, heartbeat=datetime.datetime.utcnow(), where={"status": "reserved"}
        )

    def fetch_trials_by_status(self, experiment, status):
        """See :func:`orion.storage.base.BaseStorageProtocol.fetch_trials_by_status`"""
        query = dict(experiment=experiment._id, status=status)
        return self._fetch_trials(query)



"""Tests the minimalist example script on scitkit-learn and its integration to Oríon."""
import os
import subprocess

import pytest

import orion.core.cli
from orion.client import create_experiment
from orion.storage.base import get_storage


def test_script_integrity(capsys):
    """Verifies the example script can run in standalone via `python ...`."""
    script = os.path.abspath("examples/scikitlearn-iris/main.py")

    return_code = subprocess.call(["python", script, "0.1"])

    assert return_code != 2, "The example script does not exists."
    assert return_code != 1, "The example script did not terminates its execution."
    assert (
        return_code == 0 and not capsys.readouterr().err
    ), "The example script encountered an error during its execution."


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_orion_runs_script(monkeypatch):
    """Verifies Oríon can execute the example script."""
    script = os.path.abspath("examples/scikitlearn-iris/main.py")
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    config = "orion_config.yaml"

    orion.core.cli.main(
        ["hunt", "--config", config, "python", script, "orion~choices([0.1])"]
    )

    experiment = create_experiment(name="scikit-iris-tutorial")
    assert experiment is not None
    assert experiment.version == 1

    keys = experiment.space.keys()
    assert len(keys) == 1
    assert "/_pos_2" in keys

    storage = get_storage()
    trials = storage.fetch_trials(uid=experiment.id)
    assert len(trials) == 1

    trial = trials[0]
    assert trial.status == "completed"
    assert trial.params["/_pos_2"] == 0.1


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_result_reproducibility(monkeypatch):
    """Verifies the script results stays consistent (with respect to the documentation)."""
    script = os.path.abspath("examples/scikitlearn-iris/main.py")
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    config = "orion_config.yaml"

    orion.core.cli.main(
        ["hunt", "--config", config, "python", script, "orion~choices([0.1])"]
    )

    experiment = create_experiment(name="scikit-iris-tutorial")
    assert "best_evaluation" in experiment.stats
    assert experiment.stats["best_evaluation"] == 0.6666666666666667



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Simple script simulating behaviour of a program that requires inputs
to determine and interact with a unique directory in local disk.
"""
import argparse
import os

from orion.client import report_results


def function(x):
    """Evaluate partial information of a quadratic."""
    z = x - 34.56789168765984988448213179176
    return 4 * z ** 2 + 23.4, 8 * z


def execute():
    """Execute a simple pipeline as an example."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-x", type=float, required=True)
    parser.add_argument("--dir", type=str, required=True)
    parser.add_argument("--name", type=str, required=True)
    parser.add_argument("--other-name", type=str, required=True)
    inputs = parser.parse_args()

    # That's what is expected to happen
    os.makedirs(
        os.path.join(inputs.dir, inputs.other_name, "my-exp-{}".format(inputs.name)),
        exist_ok=False,
    )  # Raise OSError if it exists

    y, dy = function(inputs.x)

    results = list()
    results.append(dict(name="example_objective", type="objective", value=y))
    results.append(dict(name="example_gradient", type="gradient", value=[dy]))

    report_results(results)


if __name__ == "__main__":
    execute()



# -*- coding: utf-8 -*-
"""
Pickled Database
================

Implement permanent version of :class:`orion.core.io.database.ephemeraldb.EphemeralDB`.

"""

import logging
import os
import pickle
from contextlib import contextmanager
from pickle import PicklingError

import psutil
from filelock import FileLock, SoftFileLock, Timeout

import orion.core
from orion.core.io.database import AbstractDB, DatabaseTimeout
from orion.core.io.database.ephemeraldb import EphemeralDB

log = logging.getLogger(__name__)

DEFAULT_HOST = os.path.join(orion.core.DIRS.user_data_dir, "orion", "orion_db.pkl")

TIMEOUT_ERROR_MESSAGE = """\
Could not acquire lock for PickledDB after {} seconds.

This is likely due to one or many of the following scenarios:

1. There is a large amount of workers and many simultaneous queries. This typically occurs
   when the task to optimize is short (few minutes). Try to reduce the amount of workers
   at least below 50.

2. The database is growing large with thousands of trials and many experiments.
   If so, you can use a different PickleDB (different file, that is, different `host`)
   for each experiment seperately to alleviate this issue.

3. The filesystem is slow. Parallel filesystems on HPC often suffer from
   large pool of users generating frequent I/O. In this case try using a separate
   partition that may be less affected.

If you cannot solve the issues listed above that are causing timeouts, you
may need to setup the MongoDB backend for better performance.
See https://orion.readthedocs.io/en/stable/install/database.html
"""


def find_unpickable_doc(dict_of_dict):
    """Look for a dictionary that cannot be pickled."""
    for name, collection in dict_of_dict.items():
        documents = collection.find()

        for doc in documents:
            try:
                pickle.dumps(doc)

            except (PicklingError, AttributeError):
                return name, doc

    return None, None


def find_unpickable_field(doc):
    """Look for a field in a dictionary that cannot be pickled"""
    if not isinstance(doc, dict):
        doc = doc.to_dict()

    for k, v in doc.items():
        try:
            pickle.dumps(v)

        except (PicklingError, AttributeError):
            return k, v

    return None, None


# pylint: disable=too-many-public-methods
class PickledDB(AbstractDB):
    """Pickled EphemeralDB to support permanancy and concurrency

    This is a very simple and inefficient implementation of a permanent database on disk for Oríon.
    The data is loaded from disk for every operation, and every operation is protected with a
    filelock.

    Parameters
    ----------
    host: str
        File path to save pickled ephemeraldb.  Default is {user data dir}/orion/orion_db.pkl ex:
        $HOME/.local/share/orion/orion_db.pkl
    timeout: int
        Maximum number of seconds to wait for the lock before raising DatabaseTimeout.
        Default is 60.

    """

    # pylint: disable=unused-argument
    def __init__(self, host="", timeout=60, *args, **kwargs):
        if host == "":
            host = DEFAULT_HOST
        super(PickledDB, self).__init__(host)

        self.host = os.path.abspath(host)

        self.timeout = timeout

        if os.path.dirname(host):
            os.makedirs(os.path.dirname(host), exist_ok=True)

    @property
    def is_connected(self):
        """Return true, always."""
        return True

    def initiate_connection(self):
        """Do nothing"""
        pass

    def close_connection(self):
        """Do nothing"""
        pass

    def ensure_index(self, collection_name, keys, unique=False):
        """Create given indexes if they do not already exist in database.

        Indexes are only created if `unique` is True.
        """
        with self.locked_database() as database:
            database.ensure_index(collection_name, keys, unique=unique)

    def index_information(self, collection_name):
        """Return dict of names and sorting order of indexes"""
        with self.locked_database(write=False) as database:
            return database.index_information(collection_name)

    def drop_index(self, collection_name, name):
        """Remove index from the database"""
        with self.locked_database() as database:
            return database.drop_index(collection_name, name)

    def write(self, collection_name, data, query=None):
        """Write new information to a collection. Perform insert or update.

        .. seealso:: :meth:`orion.core.io.database.AbstractDB.write` for argument documentation.

        """
        with self.locked_database() as database:
            return database.write(collection_name, data, query=query)

    def read(self, collection_name, query=None, selection=None):
        """Read a collection and return a value according to the query.

        .. seealso:: :meth:`orion.core.io.database.AbstractDB.read` for argument documentation.

        """
        with self.locked_database(write=False) as database:
            return database.read(collection_name, query=query, selection=selection)

    def read_and_write(self, collection_name, query, data, selection=None):
        """Read a collection's document and update the found document.

        Returns the updated document, or None if nothing found.

        .. seealso:: :meth:`orion.core.io.database.AbstractDB.read_and_write` for
                     argument documentation.

        """
        with self.locked_database() as database:
            return database.read_and_write(
                collection_name, query=query, data=data, selection=selection
            )

    def count(self, collection_name, query=None):
        """Count the number of documents in a collection which match the `query`.

        .. seealso:: :meth:`orion.core.io.database.AbstractDB.count` for argument documentation.

        """
        with self.locked_database(write=False) as database:
            return database.count(collection_name, query=query)

    def remove(self, collection_name, query):
        """Delete from a collection document[s] which match the `query`.

        .. seealso:: :meth:`orion.core.io.database.AbstractDB.remove` for argument documentation.

        """
        with self.locked_database() as database:
            return database.remove(collection_name, query=query)

    def _get_database(self):
        """Read fresh DB state from pickled file"""
        if not os.path.exists(self.host):
            return EphemeralDB()

        with open(self.host, "rb") as f:
            data = f.read()
            if not data:
                database = EphemeralDB()
            else:
                database = pickle.loads(data)

        return database

    def _dump_database(self, database):
        """Write pickled DB on disk"""
        tmp_file = self.host + ".tmp"

        try:
            with open(tmp_file, "wb") as f:
                pickle.dump(database, f)

        except (PicklingError, AttributeError):
            # pylint: disable=protected-access
            collection, doc = find_unpickable_doc(database._db)
            log.error(
                "Document in (collection: %s) is not pickable\ndoc: %s",
                collection,
                doc.to_dict() if hasattr(doc, "to_dict") else str(doc),
            )

            key, value = find_unpickable_field(doc)
            log.error("because (value %s) in (field: %s) is not pickable", value, key)
            raise

        os.rename(tmp_file, self.host)

    @contextmanager
    def locked_database(self, write=True):
        """Lock database file during wrapped operation call."""
        lock = _create_lock(self.host + ".lock")

        try:
            with lock.acquire(timeout=self.timeout):
                database = self._get_database()

                yield database

                if write:
                    self._dump_database(database)
        except Timeout as e:
            raise DatabaseTimeout(TIMEOUT_ERROR_MESSAGE.format(self.timeout)) from e

    @classmethod
    def get_defaults(cls):
        """Get database arguments needed to create a database instance.

        .. seealso:: :meth:`orion.core.io.database.AbstractDB.get_defaults`
                     for argument documentation.

        """
        return {"host": DEFAULT_HOST}


local_file_systems = ["ext2", "ext3", "ext4", "ntfs"]


def _fs_support_globalflock(file_system):
    if file_system.fstype == "lustre":
        return ("flock" in file_system.opts) and ("localflock" not in file_system.opts)

    elif file_system.fstype == "beegfs":
        return "tuneUseGlobalFileLocks" in file_system.opts

    elif file_system.fstype == "gpfs":
        return True

    elif file_system.fstype == "nfs":
        return False

    return file_system.fstype in local_file_systems


def _find_mount_point(path):
    """Finds the mount point used to access `path`."""
    path = os.path.abspath(path)
    while not os.path.ismount(path):
        path = os.path.dirname(path)

    return path


def _get_fs(path):
    """Gets info about the filesystem on which `path` lives."""
    mount = _find_mount_point(path)

    for file_system in psutil.disk_partitions(True):
        if file_system.mountpoint == mount:
            return file_system

    return None


def _create_lock(path):
    """Create lock based on file system capabilities

    Determine if we can rely on the fcntl module for locking files.
    Otherwise, fallback on using the directory creation atomicity as a locking mechanism.
    """
    file_system = _get_fs(path)

    if _fs_support_globalflock(file_system):
        log.debug("Using flock.")
        return FileLock(path)
    else:
        log.debug("Cluster does not support flock. Falling back to softfilelock.")
        return SoftFileLock(path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a stress tests on python API."""
import os
import random
import time
import traceback
from multiprocessing import Pool

import matplotlib.pyplot as plt
from pymongo import MongoClient

from orion.client import create_experiment
from orion.core.io.database import DatabaseTimeout
from orion.core.utils.exceptions import SampleTimeout
from orion.core.utils.singleton import update_singletons

DB_FILE = "stress.pkl"


def f(x, worker):
    """Sleep and return objective equal to param"""
    print(f"{worker: 6d}   {x: 5f}")
    time.sleep(max(0, random.gauss(1, 0.2)))
    return [dict(name="objective", value=x, type="objective")]


def get_experiment(storage, space_type, size):
    """Create an experiment or load from DB if already existing

    Parameters
    ----------
    storage: str
        Can be `pickleddb` or `mongodb`. A default configuration is used for each.
    space_type: str
        Can be one of
        - `discrete` Search space is discrete and limited to `max_trials`
        - `real-seeded` Search space is continuous and algos is seeded, leading to many race
          conditions while algos are sampling the same points in parallel, or
        - `real` Search space is real and algo is not seeded, leading to very few race conditions.
    size: int
        This defines `max_trials`, and the size of the search space (`uniform(0, size)`).

    """
    if storage == "pickleddb":
        storage_config = {"type": "pickleddb", "host": DB_FILE}
    elif storage == "mongodb":
        storage_config = {
            "type": "mongodb",
            "name": "stress",
            "host": "mongodb://user:pass@localhost",
        }

    discrete = space_type == "discrete"
    high = size  # * 2

    return create_experiment(
        "stress-test",
        space={"x": f"uniform(0, {high}, discrete={discrete})"},
        max_trials=size,
        max_idle_time=60 * 5,
        algorithms={"random": {"seed": None if space_type == "real" else 1}},
        storage={"type": "legacy", "database": storage_config},
    )


def worker(worker_id, storage, space_type, size):
    """Run trials until experiment is done

    Parameters
    ----------
    worker_id: int
        ID of the worker. This is used to distinguish logs from different workers.
    storage: str
        See `get_experiment`.
    space_type: str
        See `get_experiment`.
    size: int
        See `get_experiment`.

    """
    try:
        experiment = get_experiment(storage, space_type, size)

        assert experiment.version == 1, experiment.version

        print(f"{worker_id: 6d} enters")

        num_trials = 0
        while not experiment.is_done:
            try:
                trial = experiment.suggest()
            except SampleTimeout:
                trial = None

            if trial is None:
                break

            results = f(trial.params["x"], worker_id)
            num_trials += 1
            experiment.observe(trial, results=results)

        print(f"{worker_id: 6d} leaves | is done? {experiment.is_done}")
    except DatabaseTimeout as e:
        print(f"{worker_id: 6d} timeouts and leaves")
        return num_trials
    except Exception as e:
        print(f"{worker_id: 6d} crashes")
        traceback.print_exc()
        return None

    return num_trials


def stress_test(storage, space_type, workers, size):
    """Spawn workers and run stress test with verifications

    Parameters
    ----------
    storage: str
        See `get_experiment`.
    space_type: str
        See `get_experiment`.
    workers: int
        Number of workers to run in parallel.
    size: int
        See `get_experiment`.

    Returns
    -------
    `list` of `orion.core.worker.trial.Trial`
        List of all trials at the end of the stress test

    """
    if storage == "pickleddb":
        if os.path.exists(DB_FILE):
            os.remove(DB_FILE)
    elif storage == "mongodb":
        client = MongoClient(username="user", password="pass", authSource="stress")
        database = client.stress
        database.experiments.drop()
        database.lying_trials.drop()
        database.trials.drop()
        database.workers.drop()
        database.resources.drop()
        client.close()
    update_singletons()

    print("Worker  |  Point")

    with Pool(workers) as p:
        results = p.starmap(
            worker,
            zip(
                range(workers),
                [storage] * workers,
                [space_type] * workers,
                [size] * workers,
            ),
        )

    assert (
        None not in results
    ), "A worker crashed unexpectedly. See logs for the error messages."
    assert all(n > 0 for n in results), "A worker could not execute any trial."

    if space_type in ["discrete", "real-seeded"]:
        assert sum(results) == size, results
    else:
        assert sum(results) >= size, results

    experiment = get_experiment(storage, space_type, size)

    trials = experiment.fetch_trials()

    if storage == "pickleddb":
        os.remove(DB_FILE)
    elif storage == "mongodb":
        client = MongoClient(username="user", password="pass", authSource="stress")
        database = client.stress
        database.experiments.drop()
        database.lying_trials.drop()
        database.trials.drop()
        database.workers.drop()
        database.resources.drop()
        client.close()
    update_singletons()

    return trials


def get_timestamps(trials, size, space_type):
    """Get start timestamps of the trials

    Parameters
    ----------
    trials: `list` of `orion.core.worker.trial.Trial`
        List of all trials at the end of the stress test
    space_type: str
        See `get_experiment`.
    size: int
        See `get_experiment`.

    Returns
    -------
    (`list`, `list`)
        Where rval[0] is start timestamp and rval[1] is the index of the trial.
        For instance the i-th trial timestamp is rval[0][rval[1].index(i)].

    """
    hparams = set()
    x = []
    y = []

    start_time = None
    for i, trial in enumerate(trials):
        hparams.add(trial.params["x"])
        assert trial.objective.value == trial.params["x"]
        if start_time is None:
            start_time = trial.submit_time
        x.append((trial.submit_time - start_time).total_seconds())
        y.append(i)

    if space_type in ["discrete", "real-seeded"]:
        assert len(hparams) == size
    else:
        assert len(hparams) >= size

    return x[:size], y[:size]


def benchmark(workers, size):
    """Get start timestamps of the trials

    Parameters
    ----------
    workers: int
        see: `stress_test`.
    size: int
        See `get_experiment`.

    Returns
    -------
    dict
        Dictionary containing all results of all stress tests.
        Each key is (backend, space_type). See `get_experiment` for the supported types
        of `backend`s and `space_type`s. Each values result[(backend, space_type)] is
        in the form of a (x, y) tuple, where x a the list start timestamps and y is the indexes of
        the trials. See `get_timestamps` for more details.

    """
    results = {}
    for backend in ["mongodb", "pickleddb"]:
        for space_type in ["discrete", "real", "real-seeded"]:
            trials = stress_test(backend, space_type, workers, size)
            results[(backend, space_type)] = get_timestamps(trials, size, space_type)

    return results


def main():
    """Run all stress tests and render the plot"""
    size = 500

    num_workers = [1, 4, 16, 32, 64, 128]

    fig, axis = plt.subplots(
        len(num_workers),
        1,
        figsize=(5, 1.8 * len(num_workers)),
        gridspec_kw={"hspace": 0.01, "wspace": 0},
        sharex="col",
    )

    results = {}

    for i, workers in enumerate(num_workers):

        results[workers] = benchmark(workers, size)

        for backend in ["mongodb", "pickleddb"]:
            for space_type in ["discrete", "real", "real-seeded"]:
                x, y = results[workers][(backend, space_type)]
                axis[i].plot(x, y, label=f"{backend}-{space_type}")

    for i, workers in enumerate(num_workers):
        # We pick 'pickleddb' and discrete=True as the reference for the slowest ones
        x, y = results[min(num_workers)][("pickleddb", "discrete")]
        d_x = max(x) - min(x)
        d_y = max(y) - min(y)
        if i < len(num_workers) - 1:
            axis[i].text(
                min(x) + d_x * 0.6, min(y) + d_y * 0.1, f"{workers: 3d} workers"
            )
        else:
            axis[i].text(
                min(x) + d_x * 0.6, min(y) + d_y * 0.7, f"{workers: 3d} workers"
            )

    for i in range(len(num_workers) - 1):
        axis[i].spines["top"].set_visible(False)
        axis[i].spines["right"].set_visible(False)

    axis[-1].spines["right"].set_visible(False)
    axis[-1].spines["top"].set_visible(False)

    axis[-1].set_xlabel("Time (s)")
    axis[-1].set_ylabel("Number of trials")
    axis[-1].legend()

    plt.subplots_adjust(left=0.15, bottom=0.05, top=1, right=1)

    plt.savefig("test.png")


if __name__ == "__main__":
    main()
